# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for creating, manipulating, and revealing additive-shared secrets."""

import logging
from math import ceil

import numpy

from cicada.communicator.interface import Communicator
import cicada.encoder
import cicada.additive
import cicada.shamir
import random

class ActiveArrayShare(object):
    """Stores the local share of a shared secret array for the Active protocol suite.

    Parameters
    ----------
    storage: tuple(:class:`numpy.ndarray`, :class:`numpy.ndarray`), required
        Local share of a secret array.
    """
    def __init__(self, storage):
        self.storage = storage


    def __repr__(self):
        return f"cicada.active.ActiveArrayShare(storage={self._storage})" # pragma: no cover


    def __getitem__(self, index):
        return ActiveArrayShare((cicada.additive.AdditiveArrayShare(self.additive_subshare.storage[index]), cicada.shamir.ShamirArrayShare(self.shamir_subshare.storage[index])))
                

    @property
    def storage(self):
        """Local share of an shared secret array.

        Returns
        -------
        storage: :class:`object`
            Private storage for the local share of a secret array created by
            the active protocol suite.  Access is provided only for
            serialization and communication - callers must use
            :class:`ActiveProtocolSuite` to manipulate secret shares.
        """
        return self._storage

    @property
    def additive_subshare(self):
        """Local share of an shared secret array.

        Returns
        -------
        storage: :class:`object`
            Private storage for the local share of a secret array created by
            the active protocol suite.  Access is provided only for
            serialization and communication - callers must use
            :class:`ActiveProtocolSuite` to manipulate secret shares.
        """
        return self._storage[0]

    @property
    def shamir_subshare(self):
        """Local share of an shared secret array.

        Returns
        -------
        storage: :class:`object`
            Private storage for the local share of a secret array created by
            the active protocol suite.  Access is provided only for
            serialization and communication - callers must use
            :class:`ActiveProtocolSuite` to manipulate secret shares.
        """
        return self._storage[1]

    @storage.setter
    def storage(self, storage):
        if not len(storage) == 2:
            raise ValueError(f"Expected instances of AdditiveArrayShare and ShamirArrayShare, got {storage} instead.") # pragma: no cover
        if not isinstance(storage[0], cicada.additive.AdditiveArrayShare):
            raise ValueError(f"Expected instances of AdditiveArrayShare and ShamirArrayShare, got {storage} instead.") # pragma: no cover
        if not isinstance(storage[1], cicada.shamir.ShamirArrayShare):
            raise ValueError(f"Expected instances of AdditiveArrayShare and ShamirArrayShare, got {storage} instead.") # pragma: no cover
        self._storage = (storage[0], storage[1])


class ConsistencyError(Exception):
    pass

class ActiveProtocolSuite(object):
    """Protocol suite implementing computation with shared secrets that is secure against an active adversary.

    Implements "Combining Shamir & additive secret sharing to improve
    efficiency of SMC primitives against malicious adversaries" by Goss, which
    provides honest majority security with abort.

    Both :class:`~cicada.additive.AdditiveProtocolSuite` and
    :class:`~cicada.shamir.ShamirProtocolSuite` are used for the implementation.

    Note
    ----
    Creating the protocol is a collective operation that *must*
    be called by all players that are members of `communicator`.

    Parameters
    ----------
    communicator: :class:`cicada.communicator.interface.Communicator`, required
        The communicator that this protocol will use for communication.
    seed: :class:`int`, optional
        Seed used to initialize random number generators.  For privacy, this
        value should be different for each player.  By default, the seed will
        be chosen at random, and is guaranteed to be different even on forked
        processes.  If you specify `seed` yourself, the actual seed used will
        be the sum of this value and the value of `seed_offset`.
    seed_offset: :class:`int`, optional
        Value added to the value of `seed`.  This value defaults to the player's
        rank.
    modulus: :class:`int`, optional
        Field size for storing encoded values.  Defaults to the largest prime
        less than :math:`2^{64}`.
    precision: :class:`int`, optional
        The number of bits for storing fractions in encoded values.  Defaults
        to 16.
    """
    def __init__(self, communicator,*, threshold, seed=None, seed_offset=None, modulus=18446744073709551557, precision=16):
        if not isinstance(communicator, Communicator):
            raise ValueError("A Cicada communicator is required.") # pragma: no cover

        # Setup a pseudo-random sharing of zero, using code drawn from
        # https://github.com/facebookresearch/CrypTen/blob/master/crypten/__init__.py

        # Generate random seeds for Generators
        # NOTE: Chosen seed can be any number, but we choose a random 64-bit
        # integer here so other players cannot guess its value.

        # We sometimes get here from a forked process, which causes all players
        # to have the same RNG state. Reset the seed to make sure RNG streams
        # are different for all the players. We use numpy's random generator
        # here since initializing it without a seed will produce different
        # seeds even from forked processes.

        max_threshold = (communicator.world_size+1) // 2
        if threshold > max_threshold:
            min_world_size = (2 * threshold) - 1
            raise ValueError(f"threshold must be <= {max_threshold}, or world_size must be >= {min_world_size}")
        self._communicator = communicator
        self._encoder = cicada.encoder.FixedFieldEncoder(modulus=modulus, precision=precision)
        self.aprotocol = cicada.additive.AdditiveProtocolSuite(communicator=communicator, seed=seed, seed_offset=seed_offset, modulus=modulus, precision=precision)
        self.sprotocol = cicada.shamir.ShamirProtocolSuite(communicator=communicator, seed=seed, seed_offset=seed_offset, modulus=modulus, precision=precision, threshold=threshold)


    def _assert_binary_compatible(self, lhs, rhs, lhslabel, rhslabel):
        self._assert_unary_compatible(lhs, lhslabel)
        self._assert_unary_compatible(rhs, rhslabel)
        if lhs.additive_subshare.storage.shape != rhs.additive_subshare.storage.shape :
            raise ValueError(f"{lhslabel} and {rhslabel} additive shares, the additive subshares must be the same shape, got {lhs.additive_subshare.storage.shape} and {rhs.additive_subshare.storage.shape} instead.") # pragma: no cover
        if lhs.shamir_subshare.storage.shape != rhs.shamir_subshare.storage.shape:
            raise ValueError(f"{lhslabel} and {rhslabel} shamir shares, the Shamir subshares must be the same shape, got {lhs.shamir_subshare.storage.shape} and {rhs.shamir_subshare.storage.shape} instead.") # pragma: no cover


    def _assert_unary_compatible(self, share, label):
        if not isinstance(share, ActiveArrayShare):
            raise ValueError(f"{label} must be an instance of ActiveArrayShare, got {type(share)} instead.") # pragma: no cover


    def absolute(self, operand):
        """Return the elementwise absolute value of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared value to which the absolute value function should be applied.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            Secret-shared elementwise absolute value of `operand`.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((self.aprotocol.absolute(operand.additive_subshare), self.sprotocol.absolute(operand.shamir_subshare)))


    def add(self, lhs, rhs):
        """Return the elementwise sum of two secret shared arrays.

        The result is the secret shared elementwise sum of the operands.  If
        revealed, the result will need to be decoded to obtain the actual sum.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared value to be added.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared value to be added.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            Secret-shared sum of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.add(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.add(lhs.shamir_subshare, rhs.shamir_subshare)))


    def additive_inverse(self, operand):
        """Return an elementwise additive inverse of a shared array
        in the context of the underlying finite field. Explicitly, this
        function returns a same shape array which, when added
        elementwise with operand, will return a same shape array comprised
        entirely of zeros (the additive identity).

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.
        This function does not take into account any field-external symantics.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array to be additively inverted.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret additive inverse of operand on an elementwise basis.
        """
        self._assert_unary_compatible(operand, "operand")

        return ActiveArrayShare((self.aprotocol.additive_inverse(operand.additive_subshare), self.sprotocol.additive_inverse(operand.shamir_subshare)))

    def bit_compose(self, operand):
        """given an operand in a bitwise decomposed representation, compose it into shares of its field element representation.

        Note
        ----
        The operand *must* be encoded with FixedFieldEncoder.  The result will
        have one more dimension than the operand, containing the returned bits
        in big-endian order.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Shared secret to be truncated.

        Returns
        -------
        array: :class:`ActiveArrayShare`
            Share of the bit decomposed secret.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.bit_compose(operand.additive_subshare), self.sprotocol.bit_compose(operand.shamir_subshare)))

    def bit_decompose(self, operand, num_bits=None):
        """Decompose operand into shares of its bitwise representation.

        Note
        ----
        The operand *must* be encoded with FixedFieldEncoder.  The result will
        have one more dimension than the operand, containing the returned bits
        in big-endian order.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Shared secret to be truncated.

        Returns
        -------
        array: :class:`ActiveArrayShare`
            Share of the bit decomposed secret.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.bit_decompose(operand.additive_subshare, num_bits), self.sprotocol.bit_decompose(operand.shamir_subshare, num_bits)))

    def verify(self, operand):
        """Provides a means by which consistency of shares can be checked among all the parties by the use of a ZKP
        This method allows the parties to prove to one another (in zero knowledge) that a consistent set of shares is known by all parties.
        It can provide a 'safe' point at which calculations were known to be good that parties can 'rewind' to if consistency problems are 
        later discovered in the protocol.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Shared secret to have its consistency verified.

        Returns
        -------
        consistency: :class:`bool`
            True if ZKP is successful i.e., everyone has demonstrated that they know consistent shares, and False otherwise.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        a_share = operand.additive_subshare
        s_share = operand.shamir_subshare
        zero = cicada.shamir.ShamirArrayShare(self.sprotocol._encoder.subtract(s_share.storage, numpy.array((pow(self.sprotocol._revealing_coef[self.communicator.rank], self._encoder.modulus-2, self._encoder.modulus) * a_share.storage) % self._encoder.modulus, dtype=object)))
        consistency = numpy.all(self.sprotocol._reveal(zero) == numpy.zeros(zero.storage.shape))
        return consistency

    @property
    def communicator(self):
        """Return the :class:`~cicada.communicator.interface.Communicator` used by this protocol."""
        return self._communicator


    def divide(self, lhs, rhs):
        """Elementwise division of two secret shared arrays.

        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise division of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        result = self.untruncated_divide(lhs, rhs)
        result = self.truncate(result)
        return result


    def dot(self, lhs, rhs):
        """Return the dot product of two secret shared vectors.

        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared vector.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared vector.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared dot product of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        result = self.untruncated_multiply(lhs, rhs)
        result = self.sum(result)
        result = self.truncate(result)
        return result


    def equal(self, lhs, rhs):
        """Return an elementwise probabilistic equality comparison between secret shared arrays.

        The result is the secret shared elementwise comparison `lhs` == `rhs`.
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared value to be compared.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared value to be compared.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared result of computing `lhs` == `rhs` elementwise.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.equal(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.equal(lhs.shamir_subshare, rhs.shamir_subshare)))


    def floor(self, operand):
        """Return the largest integer less-than-or-equal-to `operand`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Shared secret to which floor should be applied.

        Returns
        -------
        array: :class:`ActiveArrayShare`
            Share of the floor value.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.floor(operand.additive_subshare), self.sprotocol.floor(operand.shamir_subshare)))


    def less(self, lhs, rhs):
        """Return an elementwise less-than comparison between secret shared arrays.

        The result is the secret shared elementwise comparison `lhs` < `rhs`.
        When revealed, the result will contain the values `0` or `1`, which do
        not need to be decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared value to be compared.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared value to be compared.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared result of computing `lhs` < `rhs` elementwise.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.less(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.less(lhs.shamir_subshare, rhs.shamir_subshare)))


    def less_than_zero(self, operand):
        """Return an elementwise less-than comparison between operand elements and zero.

        The result is the secret shared elementwise comparison `operand` < `0`.
        When revealed, the result will contain the values `0` or `1`, which do
        not need to be decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared value to be compared.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared result of computing `operand` < `0` elementwise.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.less_than_zero(operand.additive_subshare), self.sprotocol.less_than_zero(operand.shamir_subshare)))


    def logical_and(self, lhs, rhs):
        """Return an elementwise logical AND of two secret shared arrays.

        The operands *must* contain the *field* values `0` or `1`.  The result
        will be the secret shared elementwise logical AND of `lhs` and `rhs`.
        When revealed, the result will contain the values `0` or `1`, which do
        not need to be decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array to be AND'ed.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array to be AND'ed.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret elementwise logical AND of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.logical_and(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.logical_and(lhs.shamir_subshare, rhs.shamir_subshare)))


    def logical_not(self, operand):
        """Return an elementwise logical NOT of two secret shared array.

        The operand *must* contain the *field* values `0` or `1`.  The result
        will be the secret shared elementwise logical negation of `operand`.
        When revealed, the result will contain the values `0` or `1`, which do
        not need to be decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array to be negated.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret elementwise logical NOT of `operand`.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.logical_not(operand.additive_subshare), self.sprotocol.logical_not(operand.shamir_subshare)))


    def logical_or(self, lhs, rhs):
        """Return an elementwise logical OR of two secret shared arrays.

        The operands *must* contain the *field* values `0` or `1`.  The result
        will be the secret shared elementwise logical OR of `lhs` and `rhs`.
        When revealed, the result will contain the values `0` or `1`, which do
        not need to be decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array to be OR'd.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array to be OR'd.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret elementwise logical OR of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.logical_or(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.logical_or(lhs.shamir_subshare, rhs.shamir_subshare)))


    def logical_xor(self, lhs, rhs):
        """Return an elementwise logical exclusive OR of two secret shared arrays.

        The operands *must* contain the *field* values `0` or `1`.  The result
        will be the secret shared elementwise logical XOR of `lhs` and `rhs`.
        When revealed, the result will contain the values `0` or `1`, which do
        not need to be decoded.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array to be exclusive OR'd.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared array to be exclusive OR'd.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret elementwise logical exclusive OR of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.logical_xor(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.logical_xor(lhs.shamir_subshare, rhs.shamir_subshare)))


    def max(self, lhs, rhs):
        """Return the elementwise maximum of two secret shared arrays.

        The result is the secret shared elementwise maximum of the operands.
        If revealed, the result will need to be decoded to obtain the actual
        maximum values. Note: the field element ( if in the 'negative' range
        of the field consider only its magnitude ) should be less than
        a quarter of the modulus for this method to be accurate in general.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared operand.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared operand.

        Returns
        -------
        max: :class:`ActiveArrayShare`
            Secret-shared elementwise maximum of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.max(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.max(lhs.shamir_subshare, rhs.shamir_subshare)))


    def min(self, lhs, rhs):
        """Return the elementwise minimum of two secret shared arrays.

        The result is the secret shared elementwise minimum of the operands.
        If revealed, the result will need to be decoded to obtain the actual
        minimum values. Note: the field element ( if in the 'negative' range
        of the field consider only its magnitude ) should be less than
        a quarter of the modulus for this method to be accurate in general.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared operand.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared operand.

        Returns
        -------
        min: :class:`ActiveArrayShare`
            Secret-shared elementwise minimum of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.min(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.min(lhs.shamir_subshare, rhs.shamir_subshare)))


    def multiplicative_inverse(self, operand):
        """Return an elementwise multiplicative inverse of a shared array
        in the context of the underlying finite field. Explicitly, this
        function returns a same shape array which, when multiplied
        elementwise with operand, will return a same shape array comprised
        entirely of ones assuming operand is entirely non-trivial elements.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.
        This function does not take into account any field-external symantics.
        There is a potential for information leak here if operand contains any
        zero elements, that will be revealed. There is a small probability,
        2^-operand.storage.size, for this approach to fail by zero being randomly
        generated by the parties as the mask.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array to be multiplicatively inverted.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret multiplicative inverse of operand on an elementwise basis.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.multiplicative_inverse(operand.additive_subshare), self.sprotocol.multiplicative_inverse(operand.shamir_subshare)))


    def multiply(self, lhs, rhs):
        """Return the elementwise product of two secret shared arrays.

        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared arrays.
        rhs: :class:`ActiveArrayShare`, required
            Secret shared arrays.

        Returns
        -------
        result: :class:`ActiveArrayShare`
            Secret-shared elementwise product of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        result = self.untruncated_multiply(lhs, rhs)
        result = self.truncate(result)
        return result


    def private_public_power(self, lhs, rhspub):
        """Raise the array contained in lhs to the power rshpub on an elementwise basis

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Shared secret to which floor should be applied.
        rhspub: :class:`int`, required
            a publically known integer and the power to which each element in lhs should be raised

        Returns
        -------
        array: :class:`ActiveArrayShare`
            Share of the array elements from lhs all raised to the power rhspub.
        """
        if not isinstance(lhs, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.private_public_power(lhs.additive_subshare, rhspub), self.sprotocol.private_public_power(lhs.shamir_subshare, rhspub)))


    def private_public_power_field(self, lhs, rhspub):
        """Raise the array contained in lhs to the power rshpub on an elementwise basis

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Shared secret to which floor should be applied.
        rhspub: :class:`int`, required
            a publically known integer and the power to which each element in lhs should be raised

        Returns
        -------
        array: :class:`ActiveArrayShare`
            Share of the array elements from lhs all raised to the power rhspub.
        """
        if not isinstance(lhs, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.private_public_power_field(lhs.additive_subshare, rhspub), self.sprotocol.private_public_power_field(lhs.shamir_subshare, rhspub)))

    def private_public_subtract(self, lhs, rhs):
        """Return the elementwise difference between public and secret shared arrays.

        All players *must* supply the same value of `lhs` when calling this
        method.  The result will be the secret shared elementwise difference
        between the public (known to all players) `lhs` array and the private
        (secret shared) `rhs` array.  If revealed, the result will need to be
        decoded to obtain the actual difference.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared value from which rhs should be subtracted.
        rhs: :class:`numpy.ndarray`, required
            Public value, which must have been encoded with this protocol's
            :attr:`encoder`.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret shared difference `lhs` - `rhs`.
        """
        self._assert_unary_compatible(lhs, "lhs")

        return ActiveArrayShare((self.aprotocol.private_public_subtract(lhs.additive_subshare, rhs), self.sprotocol.private_public_subtract(lhs.shamir_subshare, rhs)))


    def random_bitwise_secret(self, *, bits, src=None, generator=None, shape=None):
        """Return a vector of randomly generated bits.

        This method is secure against non-colluding semi-honest adversaries.  A
        subset of players (by default: all) generate and secret share vectors
        of pseudo-random bits which are then xored together elementwise.
        Communication and computation costs increase with the number of bits
        and the number of players, while security increases with the number of
        players.

        Parameters
        ----------
        bits: :class:`int`, required
            Number of bits to generate.
        src: sequence of :class:`int`, optional
            Players that will contribute to random bit generation.  By default,
            all players contribute.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for sampling.  By default,
            `numpy.random.default_rng()` will be used.

        Returns
        -------
        bits: :class:`ActiveArrayShare`
            A share of the randomly-generated bits that make-up the secret.
        secret: :class:`ActiveArrayShare`
            A share of the value defined by `bits` (in big-endian order).
        """
        bs_add, ss_add = self.aprotocol.random_bitwise_secret(bits=bits, src=src, generator=generator, shape=shape)
        shamadd = []
        for i in self.communicator.ranks:
            shamadd.append(self.sprotocol._share(src=i, secret=ss_add.storage, shape=ss_add.storage.shape))
        ss_sham = cicada.shamir.ShamirArrayShare(numpy.array(sum([x.storage for x in shamadd]), dtype=self._encoder.dtype))
        shamadd = []
        for i in self.communicator.ranks:
            shamadd.append(self.sprotocol._share(src=i, secret=bs_add.storage, shape=bs_add.storage.shape))
        bs_sham = cicada.shamir.ShamirArrayShare(numpy.array(sum([x.storage for x in shamadd]), dtype=self._encoder.dtype))
        return (ActiveArrayShare((bs_add, bs_sham)), ActiveArrayShare((ss_add, ss_sham)))


    def relu(self, operand):
        """Return the elementwise ReLU of a secret shared array.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared value to which the ReLU function should be applied.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            Secret-shared elementwise ReLU of `operand`.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.relu(operand.additive_subshare), self.sprotocol.relu(operand.shamir_subshare)))

    def reshare(self, *, operand):
        """Convert a private array to an additive secret share.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`
            The local share of the secret shared array.

        Returns
        -------
        share: :class:`ActiveArrayShare`
            The local share of the secret shared array, now rerandomized.
        """

        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        reshared = ActiveArrayShare((self.aprotocol.reshare(operand=operand.additive_subshare), self.sprotocol.reshare(operand=operand.shamir_subshare)))
        if self.verify(reshared):
            return reshared
        else:
            raise ConsistencyError("Secret Shares being reshared are inconsistent")
        

    def _reveal(self, share, dst=None):
        """Reveals a secret shared value to a subset of players.

        Note
        ----
        In most cases the revealed secret needs to be decoded with this
        protocol's :attr:`encoder` to reveal the actual value.

        This is a collective operation that *must* be called by all players
        that are members of :attr:`communicator`, whether they are receiving
        the revealed secret or not.

        Parameters
        ----------
        share: :class:`ActiveArrayShare`, required
            The local share of the secret to be revealed.
        dst: sequence of :class:`int`, optional
            List of players who will receive the revealed secret.  If :any:`None`
            (the default), the secret will be revealed to all players.

        Returns
        -------
        value: :class:`numpy.ndarray` or :any:`None`
            Encoded representation of the revealed secret, if this player is a
            member of `dst`, or :any:`None`
        """
        if not isinstance(share, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(share)} instead.") # pragma: no cover
        zshare = cicada.shamir.ShamirArrayShare(self.sprotocol._encoder.subtract(share.shamir_subshare.storage, numpy.array((pow(self.sprotocol._revealing_coef[self.communicator.rank], self._encoder.modulus-2, self._encoder.modulus) * share.additive_subshare.storage) % self._encoder.modulus, dtype=self._encoder.dtype)))

        a_storage = numpy.array(self.communicator.allgather(share.additive_subshare.storage), dtype=self._encoder.dtype)
        z_storage = numpy.array(self.communicator.allgather(zshare.storage), dtype=self._encoder.dtype)
        secret = []
        revealing_coef = self.sprotocol._revealing_coef
        for index in numpy.ndindex(z_storage[0].shape):
            secret.append(sum([revealing_coef[i]*z_storage[i][index] for i in range(len(revealing_coef))]))
        rev = numpy.array([x%self._encoder.modulus for x in secret], dtype=self._encoder.dtype).reshape(share.additive_subshare.storage.shape)
        if len(rev.shape) == 0 and rev:
            raise ConsistencyError("Secret Shares are inconsistent in the first stage")
        if len(rev.shape) > 0 and numpy.any(rev):
            raise ConsistencyError("Secret Shares are inconsistent in the first stage")

        secret = []
        for index in numpy.ndindex(a_storage[0].shape):
            secret.append(sum([a_storage[i][index] for i in range(len(revealing_coef))]))
        reva = numpy.array([x%self._encoder.modulus for x in secret], dtype=self._encoder.dtype).reshape(share.additive_subshare.storage.shape)
        bs_storage=numpy.zeros(z_storage.shape, dtype=self._encoder.dtype)
        for i, c in enumerate(revealing_coef):
            bs_storage[i] =  self.sprotocol._encoder.add(z_storage[i], numpy.array((pow(self.sprotocol._revealing_coef[i], self._encoder.modulus-2, self._encoder.modulus) * a_storage[i]) % self._encoder.modulus, dtype=self._encoder.dtype))
        bs_storage %= self._encoder.modulus
        s1 = random.sample(list(self.sprotocol._indices), self.sprotocol._d+1)
        s1.sort()
        revealing_coef = self.sprotocol._lagrange_coef(s1)
        sub_secret = []
        if len(z_storage[0].shape) > 0:
            for index in numpy.ndindex(z_storage[0].shape):
                loop_acc = 0
                for i,v in enumerate(s1):
                    loop_acc += revealing_coef[i]*bs_storage[v-1][index]
                sub_secret.append(loop_acc % self._encoder.modulus)
        else:
            loop_acc = 0
            for i,v in enumerate(s1):
                loop_acc += revealing_coef[i]*bs_storage[v-1]
            sub_secret.append(loop_acc % self._encoder.modulus)

        revs = numpy.array(sub_secret, dtype=self._encoder.dtype).reshape(share.additive_subshare.storage.shape)
        s2 = random.sample(list(self.sprotocol._indices), self.sprotocol._d+1)
        s2.sort()
        while s2 == s1:
            s2 = random.sample(list(self.sprotocol._indices), self.sprotocol._d+1)
            s2.sort()
        revealing_coef = self.sprotocol._lagrange_coef(s2)
        sub_secret2 = []
        if len(z_storage[0].shape) > 0:
            for index in numpy.ndindex(z_storage[0].shape):
                loop_acc = 0
                for i,v in enumerate(s2):
                    loop_acc += revealing_coef[i]*bs_storage[v-1][index]
                sub_secret2.append(loop_acc % self._encoder.modulus)
        else:
            loop_acc = 0
            for i,v in enumerate(s2):
                loop_acc += revealing_coef[i]*bs_storage[v-1]
            sub_secret2.append(loop_acc % self._encoder.modulus)

        revs2 = numpy.array(sub_secret2, dtype=self._encoder.dtype).reshape(share.additive_subshare.storage.shape)
        if len(revs.shape) > 0 or len(revs2.shape) > 0:
            if numpy.any(revs != reva) or numpy.any(revs2 != reva):
                raise ConsistencyError("Secret Shares are inconsistent in the second stage")
        else:
            if revs != reva or revs2 != reva:
                raise ConsistencyError("Secret Shares are inconsistent in the second stage")
        return revs


    def reveal(self, share, dst=None):
        """Reveals a secret shared value to a subset of players.

        Note
        ----
        This is a collective operation that *must* be called by all players
        that are members of :attr:`communicator`, whether they are receiving
        the revealed secret or not.

        Parameters
        ----------
        share: :class:`ActiveArrayShare`, required
            The local share of the secret to be revealed.
        dst: sequence of :class:`int`, optional
            List of players who will receive the revealed secret.  If :any:`None`
            (the default), the secret will be revealed to all players.

        Returns
        -------
        value: :class:`numpy.ndarray` or :any:`None`
            The revealed secret, if this player is a member of `dst`, or :any:`None`.
        """
        return self._encoder.decode(self._reveal(share, dst=dst))


    def reveal_bits(self, share, dst=None):
        """Reveals secret shared bits to a subset of players.

        Note
        ----
        This is a collective operation that *must* be called by all players
        that are members of :attr:`communicator`, whether they are receiving
        the revealed secret or not.

        Parameters
        ----------
        share: :class:`ActiveArrayShare`, required
            The local share of the secret to be revealed.
        dst: sequence of :class:`int`, optional
            List of players who will receive the revealed secret.  If :any:`None`
            (the default), the secret will be revealed to all players.

        Returns
        -------
        value: :class:`numpy.ndarray` or :any:`None`
            The revealed secret, if this player is a member of `dst`, or :any:`None`.
        """
        return self._reveal(share, dst=dst).astype(bool).astype(numpy.uint8)


    def reveal_field(self, share, dst=None):
        """Reveals secret shared field values to a subset of players.

        Note
        ----
        This is a collective operation that *must* be called by all players
        that are members of :attr:`communicator`, whether they are receiving
        the revealed secret or not.

        Parameters
        ----------
        share: :class:`ActiveArrayShare`, required
            The local share of the secret to be revealed.
        dst: sequence of :class:`int`, optional
            List of players who will receive the revealed secret.  If :any:`None`
            (the default), the secret will be revealed to all players.

        Returns
        -------
        value: :class:`numpy.ndarray` or :any:`None`
            The revealed secret, if this player is a member of `dst`, or :any:`None`.
        """
        return self._reveal(share, dst=dst)


    def share(self, *, src, secret, shape):
        """Convert a private array to an additive secret share.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        src: :class:`int`, required
            The player providing the private array to be secret shared.
        secret: :class:`numpy.ndarray` or :any:`None`, required
            The secret array to be shared.  This value is ignored for all
            players except `src`.
        shape: :class:`tuple`, required
            The shape of the secret.  Note that the shape must be consistently
            specified by all players.

        Returns
        -------
        share: :class:`ActiveArrayShare`
            The local share of the secret shared array.
        """
        return ActiveArrayShare((self.aprotocol.share(src=src, secret=secret, shape=shape), self.sprotocol.share(src=src, secret=secret, shape=shape)))


    def share_bits(self, *, src, secret, shape):
        """Convert an array of bits to an active secret share.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        src: :class:`int`, required
            The player providing the private array to be secret shared.
        secret: :class:`numpy.ndarray` or :any:`None`, required
            The secret array to be shared.  This value is ignored for all
            players except `src`.
        shape: :class:`tuple`, required
            The shape of the secret.  Note that the shape must be consistently
            specified by all players.

        Returns
        -------
        share: :class:`ActiveArrayShare`
            The local share of the secret shared array.
        """
        return ActiveArrayShare((self.aprotocol.share_bits(src=src, secret=secret, shape=shape), self.sprotocol.share_bits(src=src, secret=secret, shape=shape)))


    def subtract(self, lhs, rhs):
        """Subtract a secret shared value from a secret shared value.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Shared value.
        rhs: :class:`ActiveArrayShare`, required
            Shared value to be subtracted.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The difference `lhs` - `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.subtract(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.subtract(lhs.shamir_subshare, rhs.shamir_subshare)))


    def sum(self, operand):
        """Return the sum of a secret shared array's elements.

        The result is the secret shared sum of the array elements.  If
        revealed, the result will need to be decoded to obtain the actual sum.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared array to be summed.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            Secret-shared sum of `operand`'s elements.
        """
        self._assert_unary_compatible(operand, "operand")
        return ActiveArrayShare((self.aprotocol.sum(operand.additive_subshare), self.sprotocol.sum(operand.shamir_subshare)))


    def truncate(self, operand, *, bits=None, src=None, generator=None):
        """Remove the `bits` least significant bits from each element in a secret shared array.

        Note
        ----
        The operand *must* be encoded with FixedFieldEncoder

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Shared secret to be truncated.
        bits: :class:`int`, optional
            Number of bits to truncate - defaults to the precision of the encoder.
        src: sequence of :class:`int`, optional
            Players who will participate in generating random bits as part of
            the truncation process.  More players increases security but
            decreases performance.  Defaults to all players.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for sampling.  By default,
            `numpy.random.default_rng()` will be used.

        Returns
        -------
        array: :class:`ActiveArrayShare`
            Share of the truncated secret.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        if bits is None:
            bits = self._encoder.precision
        tbm, tshare = self.random_bitwise_secret(bits=bits, src=src, generator=generator, shape=operand.additive_subshare.storage.shape)
        rbm, rshare = self.random_bitwise_secret(bits=self._encoder.fieldbits-bits, src=src, generator=generator, shape=operand.additive_subshare.storage.shape)
        return ActiveArrayShare((self.aprotocol.truncate(operand.additive_subshare, bits=bits, src=src, generator=generator, trunc_mask=tshare.additive_subshare, rem_mask=rshare.additive_subshare), self.sprotocol.truncate(operand.shamir_subshare, bits=bits, src=src, generator=generator, trunc_mask=tshare.shamir_subshare, rem_mask=rshare.shamir_subshare)))



    def uniform(self, *, shape=None, generator=None):
        """Return a ActiveSharedArray with the specified shape and filled with random field elements

        This method is secure against non-colluding semi-honest adversaries.  A
        subset of players (by default: all) generate and secret share vectors
        of pseudo-random field elements which are then added together
        elementwise.  Computation costs increase with the number of elements to
        generate and the number of players, while security increases with the
        number of players.

        Parameters
        ----------
        shape: :class:`tuple`, optional
            Shape of the array to populate. By default, 
            a shapeless array of one random element will be generated.
        src: sequence of :class:`int`, optional
            Players that will contribute to random array generation.  By default,
            all players contribute.
        generator: :class:`numpy.random.Generator`, optional
            A psuedorandom number generator for sampling.  By default,
            `numpy.random.default_rng()` will be used.

        Returns
        -------
        secret: :class:`ActiveArrayShare`
            A share of the random generated value.
        """
        uniadd = self.aprotocol.uniform(shape=shape, generator=generator)
        shamadd = []
        for i in self.communicator.ranks:
            shamadd.append(self.sprotocol._share(src=i, secret=uniadd.storage, shape=uniadd.storage.shape))
        unisham = cicada.shamir.ShamirArrayShare(numpy.array(sum([x.storage for x in shamadd]), dtype=self._encoder.dtype))
        return ActiveArrayShare((uniadd, unisham))

    def untruncated_multiply(self, lhs, rhs):
        """Element-wise multiplication of two shared arrays.

        The operands are assumed to be vectors or matrices and their product is
        computed on an elementwise basis. Multiplication with shared secrets and
        public scalars is implemented in the encoder.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            secret value to be multiplied.
        rhs: :class:`ActiveArrayShare`, required
            secret value to be multiplied.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret elementwise product of `lhs` and `rhs`.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        return ActiveArrayShare((self.aprotocol.untruncated_multiply(lhs.additive_subshare, rhs.additive_subshare), self.sprotocol.untruncated_multiply(lhs.shamir_subshare, rhs.shamir_subshare)))


    def untruncated_divide(self, lhs, rhs):
        """Element-wise division of private values. Note: this may have a chance to leak info is the secret contained in rhs is 
        close to or bigger than 2^precision

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array dividend.
        rhs: :class:`numpy.ndarray`, required
            Public array divisor, which must *not* be encoded.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret element-wise result of lhs / rhs.
        """
        self._assert_binary_compatible(lhs, rhs, "lhs", "rhs")
        fieldbits = self._encoder.fieldbits
        truncbits = self._encoder.precision
        tbm, tshare = self.random_bitwise_secret(bits=truncbits, shape=rhs.additive_subshare.storage.shape)

        tbm, mask1 = self.random_bitwise_secret(bits=truncbits, shape=rhs.additive_subshare.storage.shape)
        tbm, rem1 = self.random_bitwise_secret(bits=fieldbits-truncbits, shape=rhs.additive_subshare.storage.shape)
        tbm, mask2 = self.random_bitwise_secret(bits=truncbits, shape=rhs.additive_subshare.storage.shape)
        tbm, rem2 = self.random_bitwise_secret(bits=fieldbits-truncbits, shape=rhs.additive_subshare.storage.shape)
        rev = self.reveal(tshare)
        resa = self.aprotocol.untruncated_divide(lhs.additive_subshare, rhs.additive_subshare, tshare.additive_subshare, mask1.additive_subshare, rem1.additive_subshare, mask2.additive_subshare, rem2.additive_subshare)
        ress = self.sprotocol.untruncated_divide(lhs.shamir_subshare, rhs.shamir_subshare,tshare.shamir_subshare, mask1.shamir_subshare, rem1.shamir_subshare, mask2.shamir_subshare, rem2.shamir_subshare)
        return ActiveArrayShare((resa, ress))


    def untruncated_private_public_divide(self, lhs, rhs):
        """Element-wise division of private and public values.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        lhs: :class:`ActiveArrayShare`, required
            Secret shared array dividend.
        rhs: :class:`numpy.ndarray`, required
            Public array divisor, which must *not* be encoded.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            The secret element-wise result of lhs / rhs.
        """
        self._assert_unary_compatible(lhs, "lhs")
        return ActiveArrayShare((self.aprotocol.untruncated_private_public_divide(lhs.additive_subshare, rhs), self.sprotocol.untruncated_private_public_divide(lhs.shamir_subshare, rhs)))

    def zigmoid(self, operand):
        r"""Compute the elementwise zigmoid function of a secret value.

        Zigmoid is an approximation of sigmoid which is more angular and is a piecewise function much
        more efficient to compute in an MPC context:

        .. math::

            zigmoid(x) = \left\{
            \begin{array}\\
                0 & if\ x<-0.5 \\
                x+0.5 & if\ -0.5\leq x \leq 0.5 \\
                1 & if x>0.5
            \end{array}
            \right.

        Note
        ----
        This is a collective operation that *must* be called
        by all players that are members of :attr:`communicator`.

        Parameters
        ----------
        operand: :class:`ActiveArrayShare`, required
            Secret shared value to which the zigmoid function should be applied.

        Returns
        -------
        value: :class:`ActiveArrayShare`
            Secret-shared elementwise zigmoid of `operand`.
        """
        if not isinstance(operand, ActiveArrayShare):
            raise ValueError(f"Expected operand to be an instance of ActiveArrayShare, got {type(operand)} instead.") # pragma: no cover
        return ActiveArrayShare((self.aprotocol.zigmoid(operand.additive_subshare), self.sprotocol.zigmoid(operand.shamir_subshare)))



# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import socket
import urllib.parse

import numpy

from cicada.calculator import Client, PlayerError, main
from cicada.communicator import SocketCommunicator

import test

#logging.basicConfig(level=logging.INFO)


def _print_stacks(context, player=None):
    results = context.calculator.command("opstack", player=player)
    stacks = _require_success(results)
    players, results = results

    for rank, stack in zip(players, stacks):
        print(f"====== Remote stack (player {rank}) ======")
        for level, item in zip(range(len(stack)-1, -1, -1), stack):
            print(f"Level {level}: {item}")


def _require_success(results):
    players, results = results

    exceptions = {}
    for rank, result in zip(players, results):
        if isinstance(result, PlayerError):
            print(f"====== Remote traceback (player {rank}) ======\n{result.traceback}")
            exceptions[rank] = result

    if exceptions:
        raise RuntimeError(f"{len(exceptions)} players raised exceptions.")

    return results


@given(u'a calculator service with {world_size} players')
def step_impl(context, world_size):
    world_size = eval(world_size)
    addresses, processes = SocketCommunicator.run_forever(world_size=world_size, fn=main)
    context.calculator = Client(addresses)
    context.calculator_processes = processes


@given(u'a calculator service with {world_size} players using unix domain sockets')
def step_impl(context, world_size):
    world_size = eval(world_size)
    addresses, processes = SocketCommunicator.run_forever(world_size=world_size, fn=main, family="file")
    context.calculator = Client(addresses)
    context.calculator_processes = processes


@given(u'a new {name} protocol suite')
def step_impl(context, name):
    _require_success(context.calculator.command("protopush", name=name))


@given(u'public value {value}')
def step_impl(context, value):
    value = numpy.array(eval(value))
    _require_success(context.calculator.command("oppush", value=value))


@given(u'player {player} secret shares the bits {secret}')
def step_impl(context, player, secret):
    player = eval(player)
    secret = numpy.array(eval(secret))

    for rank in context.calculator.ranks:
        _require_success(context.calculator.command("oppush", value=secret if player == rank else None, player=rank))
    _require_success(context.calculator.command("protocol", subcommand="share_bits", src=player, shape=secret.shape))


@given(u'player {player} secret shares {secret}')
def step_impl(context, player, secret):
    player = eval(player)
    secret = numpy.array(eval(secret))

    for rank in context.calculator.ranks:
        _require_success(context.calculator.command("oppush", value=secret if player == rank else None, player=rank))
    _require_success(context.calculator.command("protocol", subcommand="share", src=player, shape=secret.shape))


@given(u'the players reshare the secret')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="reshare"))


@given(u'player {player} tampers with the additive portion of its ActiveArrayShare')
def step_impl(context, player):
    player = eval(player)
    _require_success(context.calculator.command("opdup", player=player))
    _require_success(context.calculator.command("share", subcommand="getstorage", player=player))
    _require_success(context.calculator.command("opsplit", player=player))
    _require_success(context.calculator.command("opswap", player=player))
    _require_success(context.calculator.command("opdup", player=player))
    _require_success(context.calculator.command("share", subcommand="getstorage", player=player))
    _require_success(context.calculator.command("oppush", value=numpy.array(1), player=player))
    _require_success(context.calculator.command("add", player=player))
    _require_success(context.calculator.command("share", subcommand="setstorage", player=player))
    _require_success(context.calculator.command("opswap", player=player))
    _require_success(context.calculator.command("opcatn", player=player, n=2))
    _require_success(context.calculator.command("share", subcommand="setstorage", player=player))


@given(u'player {player} tampers with the shamir portion of its ActiveArrayShare')
def step_impl(context, player):
    player = eval(player)
    _require_success(context.calculator.command("opdup", player=player))
    _require_success(context.calculator.command("share", subcommand="getstorage", player=player))
    _require_success(context.calculator.command("opsplit", player=player))
    _require_success(context.calculator.command("opdup", player=player))
    _require_success(context.calculator.command("share", subcommand="getstorage", player=player))
    _require_success(context.calculator.command("oppush", value=numpy.array(1), player=player))
    _require_success(context.calculator.command("add", player=player))
    _require_success(context.calculator.command("share", subcommand="setstorage", player=player))
    _require_success(context.calculator.command("opcatn", player=player, n=2))
    _require_success(context.calculator.command("share", subcommand="setstorage", player=player))


@when(u'player {player} adds {value} to the share in-place')
def step_impl(context, player, value):
    player = eval(player)
    value = numpy.array(eval(value))
    _require_success(context.calculator.command("oppush", value=value, player=player))
    _require_success(context.calculator.command("protocol", subcommand="encode", player=player))
    _require_success(context.calculator.command("protocol", subcommand="inplace_add", player=player))


@when(u'player {player} subtracts {value} from the share in-place')
def step_impl(context, player, value):
    player = eval(player)
    value = numpy.array(eval(value))
    _require_success(context.calculator.command("oppush", value=value, player=player))
    _require_success(context.calculator.command("protocol", subcommand="encode", player=player))
    _require_success(context.calculator.command("protocol", subcommand="inplace_subtract", player=player))


@when(u'the players add the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="add"))


@when(u'the players compare the shares for equality')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="equal"))


@when(u'the players compare the shares with less than')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="less"))


@when(u'the players compare the shares with less than zero')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="less_than_zero"))


@when(u'the players compute the absolute value of the share')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="absolute"))


@when(u'the players compute the additive inverse')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="additive_inverse"))


@when(u'the players compute the composition of the shared bits')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="bit_compose"))


@when(u'the players compute the decomposition of the shared secrets')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="bit_decompose"))


@when(u'the players compute the dot product of the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="dot"))


@when(u'the players compute the floor of the share')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="floor"))


@when(u'the players compute the logical and of the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="logical_and"))


@when(u'the players compute the logical exclusive or of the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="logical_xor"))


@when(u'the players compute the logical not of the share')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="logical_not"))


@when(u'the players compute the logical or of the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="logical_or"))


@when(u'the players compute the maximum of the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="max"))


@when(u'the players compute the minimum of the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="min"))


@when(u'the players compute the multiplicative inverse')
def step_impl(context):
    _require_success(context.calculator.command("opdup"))
    _require_success(context.calculator.command("protocol", subcommand="multiplicative_inverse"))


@when(u'the players compute the relu of the share')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="relu"))


@when(u'the players compute the sum of the share')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="sum"))


@when(u'the players compute the zigmoid of the share')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="zigmoid"))


@when(u'the players divide the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="divide"))


@given(u'the players extract the share storage')
def step_impl(context):
    _require_success(context.calculator.command("share", subcommand="getstorage"))


@when(u'the players generate {bits} random bits')
def step_impl(context, bits):
    bits = eval(bits)
    _require_success(context.calculator.command("oppush", value=bits))
    _require_success(context.calculator.command("protocol", subcommand="random_bitwise_secret"))


@when(u'the players generate a private uniform array with shape {shape}')
def step_impl(context, shape):
    shape = eval(shape)
    _require_success(context.calculator.command("protocol", subcommand="uniform", shape=shape))


@when(u'the players multiply the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="multiply"))


@when(u'the players multiply the shares without truncation')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="untruncated_multiply"))


@when(u'the players raise the share to the public power')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="private_public_power"))


@when(u'the players raise the share to the public power in the field')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="private_public_power_field"))


@when(u'the players reveal the secret')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="reveal"))


@when(u'the players reveal the secret bits')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="reveal_bits"))


@when(u'the players reveal the field values')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="reveal_field"))


@when(u'the players subtract the shares')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="subtract"))


@when(u'the players swap')
def step_impl(context):
    _require_success(context.calculator.command("opswap"))


@when(u'the players verify the share')
def step_impl(context):
    _require_success(context.calculator.command("protocol", subcommand="verify"))


@then(u'{count} {name} protocol objects can be created without error')
def step_impl(context, count, name):
    count = eval(count)
    for index in range(count):
        _require_success(context.calculator.command("protopush", name=name))


@then(u'the players can retrieve a complete copy of the operand stack')
def step_impl(context):
    context.opstack = _require_success(context.calculator.command("opstack"))


@then(u'the stack should match {stack} for all players')
def step_impl(context, stack):
    stack = eval(stack)
    for playerstack in context.opstack:
        test.assert_equal(len(stack), len(playerstack))
        for lhs, rhs in zip(stack, playerstack):
            numpy.testing.assert_array_equal(lhs, rhs)


@then(u'the result should match {rhs} to within {digits} digits')
def step_impl(context, rhs, digits):
    rhs = eval(rhs)
    digits = eval(digits)

    for lhs in _require_success(context.calculator.command("opget")):
        numpy.testing.assert_array_almost_equal(lhs, rhs, decimal=digits)

@then(u'the results should match {rhs} to within {digits} digits')
def step_impl(context, rhs, digits):
    rhs = eval(rhs)
    digits = eval(digits)

    for lhs in _require_success(context.calculator.command("opget")):
        print(lhs)
        numpy.testing.assert_array_almost_equal(lhs, rhs, decimal=digits)

@then(u'the result should match {rhs}')
def step_impl(context, rhs):
    rhs = eval(rhs)

    for lhs in _require_success(context.calculator.command("opget")):
        numpy.testing.assert_array_equal(lhs, rhs)


@then(u'the two values should not be equal')
def step_impl(context):
    values = _require_success(context.calculator.command("opgetn", n=2))
    for lhs, rhs in values:
        test.assert_false(numpy.array_equal(lhs, rhs))


@then(u'the value of the bits in big-endian order should match the random value.')
def step_impl(context):
    values = _require_success(context.calculator.command("oppop"))
    bits = _require_success(context.calculator.command("oppop"))

    for bits, value in zip(bits, values):
        test.assert_equal(value, numpy.sum(numpy.power(2, numpy.arange(len(bits))[::-1]) * bits))


@when(u'the players raise {exception}')
def step_impl(context, exception):
    exception = eval(exception)
    players, results = context.calculator.command("raise", exception=exception)
    context.errors = results


@then(u'the returned exceptions should match {exception}')
def step_impl(context, exception):
    exception = eval(exception)
    for error in context.errors:
        test.assert_is(type(error.exception), type(exception))
        test.assert_equal(error.exception.args, exception.args)


@then(u'the results should match shape {shape}')
def step_impl(context, shape):
    shape = eval(shape)
    results = _require_success(context.calculator.command("opget"))
    for result in results:
        test.assert_equal(shape, result.shape)





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cProfile import Profile
import pstats
import time

import numpy

from cicada.additive import AdditiveProtocolSuite
from cicada.shamir import ShamirProtocolSuite
from cicada.communicator import SocketCommunicator


def main(communicator, n):
    with Profile() as profile:
        proto = ShamirProtocolSuite(communicator, threshold=2)

        generator = numpy.random.default_rng()
        a = generator.uniform(size=n) if communicator.rank==0 else None
        b = generator.uniform(size=n) if communicator.rank==1 else None

        ashare = proto.share(src=0, secret=a, shape=n)
        bshare = proto.share(src=1, secret=b, shape=n)
        cshare = proto.dot(ashare, bshare)
        c = proto.reveal(cshare)

        if communicator.rank == 0:
            stats = pstats.Stats(profile)
            stats.sort_stats(pstats.SortKey.CUMULATIVE)
            stats.print_stats(30)

SocketCommunicator.run(world_size=3, fn=main, args=[100000], timeout=100)




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements a client and a server for an MPC-as-a-service calculator."""

import logging
import pickle
import socket
import traceback
import urllib

import numpy

from cicada import Logger


class Client(object):
    """Client for working with the privacy-preserving RPN calculator service.

    Parameters
    ----------
    addresses: sequence of :class:`str`, required
        Socket addresses of the players providing an RPN calculator service.
    """
    def __init__(self, addresses):
        self._addresses = addresses

    def command(self, command, *, player=None, **kwargs):
        """Sends a command to the players providing an RPN calculator service.

        Parameters
        ----------
        command: :class:`str`
            The command to be executed by the RPN calculator service.
        player: :class:`int` or sequence of :class:`int`, optional
            The player(s) that will execute the command.  By default, the
            command will be sent to all players.
        kwargs:
            Keyword arguments that will be included with the command.

        Returns
        -------
        results: sequence of :class:`object`
            Contains the result value returned by each `player`.
        """
        if player is None:
            player = self.ranks
        if not isinstance(player, list):
            player = [player]
        players = player

        # Establish connections
        sockets = []
        for player in players:
            address = self._addresses[player]
            address = urllib.parse.urlparse(address)
            if address.scheme == "file":
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                sock.connect(address.path)
                sockets.append(sock)
            elif address.scheme == "tcp":
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((address.hostname, address.port))
                sockets.append(sock)

        # Send commands
        for sock in sockets:
            sock.sendall(pickle.dumps((command, kwargs)))

        # Receive results
        results = []
        for sock in sockets:
            result = b""
            while True:
                data = sock.recv(4096)
                if not data:
                    break
                result += data
            results.append(pickle.loads(result))

        return players, results


    @property
    def ranks(self):
        """Returns a :class:`list` of player ranks provided by the service."""
        return list(range(len(self._addresses)))


class PlayerError(Exception):
    """Returned from the RPN calculator service when a player raises an exception."""
    def __init__(self, exception, traceback):
        self.exception = exception
        self.traceback = traceback


def _send_result(client, result=None):
    client.sendall(pickle.dumps(result))
    client.close()


def main(listen_socket, communicator):
    """Implements a privacy-preserving RPN calculator service.

    This is an example of MPC-as-a-service, for cases where that is
    desirable.  Run this function using::

        SocketCommunicator.run_forever(world_size=n, fn=cicada.calculator.main)

    ... which will start the service and return a set of player addresses and
    processes.  Client code can use those addressess to connect to the service
    and issue commands, which will be executed by the service players.

    The service implements an RPN calculator where communicators, protocols,
    and operands are pushed and popped from stacks.
    """
    listen_socket.setblocking(True)
    log = Logger(logger=logging.getLogger(), communicator=communicator, sync=False)

    communicator_stack = [communicator]
    protocol_stack = []
    operand_stack = []

    while True:
        try:
            # Wait for a client to connect and send a command.
            client, addr = listen_socket.accept()
            command, kwargs = pickle.loads(client.recv(4096))

            pretty_args = ", ".join([f"{key}={value!r}" for key, value in kwargs.items()])
            pretty_command = f"{command}({pretty_args})"
            log.debug(f"Player {communicator.rank} received command: {pretty_command}")


            #############################################################
            # Commands related to the MPC service.

            # Exit the service immediately.
            if command == "quit":
                _send_result(client)
                break

            # Raise an exception for testing.
            elif command == "raise":
                raise kwargs["exception"]

            #############################################################
            # Commands related to the communicator stack.

            # Placeholder for operations to split, shrink, or connect
            # new communicators.

            #############################################################
            # Commands related to the protocol stack.

            # Push a new ActiveProtocolSuite object onto the protocol stack.
            elif command == "protopush" and kwargs["name"] == "Active":
                import cicada.active
                protocol_stack.append(cicada.active.ActiveProtocolSuite(communicator_stack[-1], threshold=2))
                _send_result(client)

            # Push a new AdditiveProtocolSuite onto the protocol stack.
            elif command == "protopush" and kwargs["name"] == "Additive":
                from cicada.additive import AdditiveProtocolSuite
                protocol_stack.append(AdditiveProtocolSuite(communicator_stack[-1]))
                _send_result(client)

            # Push a new ShamirProtocolSuite object onto the protocol stack.
            elif command == "protopush" and kwargs["name"] == "Shamir":
                from cicada.shamir import ShamirProtocolSuite
                protocol_stack.append(ShamirProtocolSuite(communicator_stack[-1], threshold=2))
                _send_result(client)

            #############################################################
            # Commands related to the operand stack.

            # Combine the top n values on the operand stack into a single operand.
            elif command == "opcatn":
                value = [operand_stack.pop() for index in range(kwargs["n"])]
                value.reverse()
                operand_stack.append(value)
                _send_result(client)

            # Duplicate the value on top of the operand stack.
            elif command == "opdup":
                value = operand_stack[-1]
                operand_stack.append(value)
                _send_result(client)

            # Get the value on top of the operand stack.
            elif command == "opget":
                _send_result(client, operand_stack[-1])

            # Get n values from the top of the operand stack.
            elif command == "opgetn":
                n = kwargs["n"]
                _send_result(client, operand_stack[-n:])

            # Pop a value from the top of the operand stack.
            elif command == "oppop":
                value = operand_stack.pop()
                _send_result(client, value)

            # Push a new value on top of the operand stack.
            elif command == "oppush":
                operand = kwargs["value"]
                operand_stack.append(operand)
                _send_result(client)

            # Split a sequence value on the top of the operand stack into separate stack values.
            elif command == "opsplit":
                value = operand_stack.pop()
                for item in value:
                    operand_stack.append(item)
                _send_result(client)

            # Return a copy of the entire operand stack.
            elif command == "opstack":
                _send_result(client, operand_stack)

            # Swap the top two values on the operand stack
            elif command == "opswap":
                operand_stack[-1], operand_stack[-2] = operand_stack[-2], operand_stack[-1]
                _send_result(client)

            #############################################################
            # Commands related to "normal" arithmetic.

            elif command == "add":
                operand_stack.append(operand_stack.pop() + operand_stack.pop())
                _send_result(client)

            #############################################################
            # Commands related to secret shares.

            # Extract the storage from a secret share on top of the operand stack.
            elif command == "share" and kwargs["subcommand"] == "getstorage":
                share = operand_stack.pop()
                operand_stack.append(share.storage)
                _send_result(client)

            # Replace the storage for a secret share on top of the operand stack.
            elif command == "share" and kwargs["subcommand"] == "setstorage":
                storage = operand_stack.pop()
                operand_stack[-1].storage = storage
                _send_result(client)

            #############################################################
            # Commands related to protocol suites.

            # Encode the value on top of the operand stack.
            elif command == "protocol" and kwargs["subcommand"] == "encode":
                protocol = protocol_stack[-1]
                value = operand_stack.pop()
                value = protocol.encoder.encode(value)
                operand_stack.append(value)
                _send_result(client)

            # Secret share a value from the top of the operand stack.
            elif command == "protocol" and kwargs["subcommand"] == "reshare":
                protocol = protocol_stack[-1]
                secret = operand_stack.pop()
                share = protocol.reshare(operand = secret)
                operand_stack.append(share)
                _send_result(client)

            # Reveal a secret shared value from the top of the operand stack.
            elif command == "protocol" and kwargs["subcommand"] == "reveal":
                protocol = protocol_stack[-1]
                share = operand_stack.pop()
                secret = protocol.reveal(share)
                operand_stack.append(secret)
                _send_result(client)

            # Reveal a secret value containing only zeros and ones from the top of the operand stack.
            elif command == "protocol" and kwargs["subcommand"] == "reveal_bits":
                protocol = protocol_stack[-1]
                share = operand_stack.pop()
                secret = protocol.reveal_bits(share)
                operand_stack.append(secret)
                _send_result(client)

            # Reveal a secret field value from the top of the operand stack without decoding.
            elif command == "protocol" and kwargs["subcommand"] == "reveal_field":
                protocol = protocol_stack[-1]
                share = operand_stack.pop()
                secret = protocol.reveal_field(share)
                operand_stack.append(secret)
                _send_result(client)

            # Secret share a value from the top of the operand stack.
            elif command == "protocol" and kwargs["subcommand"] == "share":
                src = kwargs["src"]
                shape = kwargs["shape"]
                protocol = protocol_stack[-1]
                secret = operand_stack.pop()
                share = protocol.share(src=src, secret=secret, shape=shape)
                operand_stack.append(share)
                _send_result(client)

            # Secret share bits from the top of the operand stack.
            elif command == "protocol" and kwargs["subcommand"] == "share_bits":
                src = kwargs["src"]
                shape = kwargs["shape"]
                protocol = protocol_stack[-1]
                secret = operand_stack.pop()
                share = protocol.share_bits(src=src, secret=secret, shape=shape)
                operand_stack.append(share)
                _send_result(client)

            # Unary protocol suite operations.
            elif command == "protocol" and kwargs["subcommand"] in ["absolute", "additive_inverse", "bit_compose", "bit_decompose", "floor", "logical_not", "less_than_zero", "multiplicative_inverse", "relu", "sum", "truncate", "verify", "zigmoid"]:
                protocol = protocol_stack[-1]
                a = operand_stack.pop()
                result = getattr(protocol, kwargs["subcommand"])(a)
                operand_stack.append(result)
                _send_result(client)

            # Binary protocol suite operations.
            elif command == "protocol" and kwargs["subcommand"] in ["add", "divide", "dot", "equal", "less", "logical_and", "logical_or", "logical_xor", "max", "min", "multiply", "private_public_power", "private_public_power_field", "subtract", "untruncated_divide", "untruncated_multiply"]:
                protocol = protocol_stack[-1]
                b = operand_stack.pop()
                a = operand_stack.pop()
                share = getattr(protocol, kwargs["subcommand"])(a, b)
                operand_stack.append(share)
                _send_result(client)

            # Random bitwise secret.
            elif command == "protocol" and kwargs["subcommand"] == "random_bitwise_secret":
                protocol = protocol_stack[-1]
                bits = operand_stack.pop()
                bits, secret = protocol.random_bitwise_secret(bits=bits)
                operand_stack.append(bits)
                operand_stack.append(secret)
                _send_result(client)

            # Uniform random secret generation.
            elif command == "protocol" and kwargs["subcommand"] == "uniform":
                protocol = protocol_stack[-1]
                shape = kwargs["shape"]
                secret = protocol.uniform(shape=shape)
                operand_stack.append(secret)
                _send_result(client)

            # Inplace addition.
            elif command == "protocol" and kwargs["subcommand"] == "inplace_add":
                protocol = protocol_stack[-1]
                b = operand_stack.pop()
                a = operand_stack.pop()
                protocol.encoder.inplace_add(a.storage, b)
                operand_stack.append(a)
                _send_result(client)

            # Inplace subtraction.
            elif command == "protocol" and kwargs["subcommand"] == "inplace_subtract":
                protocol = protocol_stack[-1]
                b = operand_stack.pop()
                a = operand_stack.pop()
                protocol.encoder.inplace_subtract(a.storage, b)
                operand_stack.append(a)
                _send_result(client)

            # Unknown command.
            else: # pragma: no cover
                raise ValueError(f"Unknown command {pretty_command}")

        # Something went wrong.
        except Exception as e: # pragma: no cover
            _send_result(client, PlayerError(e, traceback.format_exc()))




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy
import time
import cicada.communicator
import cicada.active
from copy import deepcopy

logging.basicConfig(level=logging.INFO)

dumb_change = 0

smart_change = not dumb_change 

def main(communicator):
    log = cicada.Logger(logging.getLogger(), communicator)
    protocol = cicada.active.ActiveProtocolSuite(communicator, threshold=3)

    # Player 0 will provide a secret which is a scalar.
    secret = numpy.array([numpy.pi,1,-2,3]) if communicator.rank == 0 else None
    log.info(f"Player {communicator.rank} secret: {secret}")

    # Create shares for the secret.
    share = protocol.share(src=0, secret=secret, shape=(4,))
    #log.info(f"Player {communicator.rank} share: {share}")

    modulus = 2**64-59
    #log.info(f"Player {communicator.rank} share consistency check: {protocol.check_commit(share)}")
    log.info(f"Player {communicator.rank} share reveal check: {protocol.reveal(share)}")
    log.info(f"Player {communicator.rank} Entering Malicious activity, since we're messing with things in a {['dumb', 'smart'][smart_change]} way,\nthe consistency check should fail in the {['first', 'second'][smart_change]} step.", src=0)
    bad_share = deepcopy(share)
    if protocol.communicator.rank == 3 and dumb_change:
        bad_share[0].storage[0] += 1
    if protocol.communicator.rank == 0 and smart_change:
        bad_share[0].storage[1] += 1
        bad_share[1].storage[1] = (bad_share[1].storage[1] + pow(5, modulus-2, modulus)) % modulus
    try:
        log.info(f"Player {communicator.rank} share consistency check - should be True: {protocol.verify(bad_share)}")
    except cicada.active.ConsistencyError as e:
        print(f'Malicious alteration detected: {e}')
    try:
        log.info(f"Player {communicator.rank} share reveal check: {protocol.reveal(bad_share)}")
    except cicada.active.ConsistencyError as e:
        log.info(f'Malicious alteration detected: {e}')



    log.info("\n\nLet's try some operations...", src=0)
    double_share = protocol.add(share, share)
    try:
        log.info(f"Player {communicator.rank} share consistency check - should be True: {protocol.verify(double_share)}")
    except cicada.active.ConsistencyError as e:
        print(f'Malicious alteration detected: {e}')
    try:
        log.info(f"Player {communicator.rank} double share reveal check: {protocol.reveal(double_share)}")
    except cicada.active.ConsistencyError as e:
        log.info(f'Malicious alteration detected: {e}')



    square_share = protocol.untruncated_multiply(share, share)
    square_share = protocol.truncate(square_share)
    try:
        log.info(f"Player {communicator.rank} share consistency check - should be True: {protocol.verify(square_share)}")
    except cicada.active.ConsistencyError as e:
        print(f'Malicious alteration detected: {e}')
    try:
        log.info(f"Player {communicator.rank} share reveal check: {protocol.reveal(square_share)}")
    except cicada.active.ConsistencyError as e:
        log.info(f'Malicious alteration detected: {e}')




    log.info(f"\n\nFinally we'll mess with the results of the previous computation now, since we're messing with things in a {['dumb', 'smart'][smart_change]} way,\nthe consistency check should fail in the {['first', 'second'][smart_change]} step.", src=0)
    if protocol.communicator.rank == 3 and dumb_change:
        square_share[0].storage[0] += 1
    if protocol.communicator.rank == 0 and smart_change:
        square_share[0].storage[1] += 1
        square_share[1].storage[1] = (square_share[1].storage[1] + pow(5, modulus-2, modulus)) % modulus
    try:
        log.info(f"Player {communicator.rank} share consistency check - should be True: {protocol.verify(square_share)}")
    except cicada.active.ConsistencyError as e:
        print(f'Malicious alteration detected: {e}')
    try:
        log.info(f"Player {communicator.rank} share reveal check: {protocol.reveal(square_share)}")
    except cicada.active.ConsistencyError as e:
        log.info(f'Malicious alteration detected: {e}')

cicada.communicator.SocketCommunicator.run(world_size=5, fn=main)




# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines abstract interfaces for network communication."""

from abc import ABCMeta, abstractmethod
import enum

class Communicator(metaclass=ABCMeta):
    """Abstract base class for objects that manage collective communications for secure multiparty computation.
    """
    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()
        return False


    @abstractmethod
    def allgather(self, value):
        """All-to-all communication.

        Note
        ----
        This method is a collective operation that *must* be called
        by all players that are members of the communicator.

        Parameters
        ----------
        value: Any picklable :class:`object`, required
            Local value that will be sent to all players.

        Returns
        -------
        values: sequence of :class:`object`
            Collection of objects gathered from every player, in rank order.
        """
        pass # pragma: no cover

    @abstractmethod
    def barrier(self):
        """Block the local process until all players have entered the barrier.

        Note
        ----
        This method is a collective operation that *must* be called
        by all players that are members of the communicator.
        """
        pass # pragma: no cover

    @abstractmethod
    def broadcast(self, *, src, value):
        """One-to-all communication.

        The `src` player broadcasts a single object to all players.

        Note
        ----
        This method is a collective operation that *must* be called
        by all players that are members of the communicator.

        Parameters
        ----------
        src: :class:`int`, required
            Rank of the player who is broadcasting.
        value: Any picklable :class:`object` or `None`, required
            Value to be broadcast by `src`.  Ignored for all other players.

        Returns
        -------
        value: :class:`object`
            The broadcast value.
        """
        pass # pragma: no cover

    @abstractmethod
    def free(self):
        """Free the communicator.

        This should be called if the communicator is no longer needed so that
        resources can be freed.  Note that communicators cannot be reused after
        they have been freed, a new communicator must be created instead.
        """
        pass # pragma: no cover

    @abstractmethod
    def gather(self, *, value, dst):
        """All-to-one communication.

        Every player sends a value to `dst`.

        Note
        ----
        This method is a collective operation that *must* be called
        by all players that are members of the communicator.

        Parameters
        ----------
        value: Any picklable :class:`object`, required
            Value to be sent to `dst`.
        dst: :class:`int`, required
            Rank of the player who will receive all of the values.

        Returns
        -------
        values: sequence of :class:`object` or None
            For the destination player, a sequence of `world_size` objects
            received from every player in rank order.  For all other players,
            `None`.
        """
        pass # pragma: no cover

    @abstractmethod
    def gatherv(self, *, src, value, dst):
        """Many-to-one communication.

        A subset of players each sends a value to `dst`.

        Note
        ----
        This method is a collective operation that *must* be called
        by all players that are members of the communicator.

        Parameters
        ----------
        src: sequence of :class:`int`, required
            Rank of each player sending a value.
        value: Any picklable :class:`object`, or :any:`None`, required
            Value to be sent to `dst`.
        dst: :class:`int`, required
            Rank of the player who will receive all of the values.

        Returns
        -------
        values: sequence of :class:`object` or None
            For the destination player, the sequence of values received from
            the `src` players in the same order as `src`.  For all other
            players, :any:`None`.
        """
        pass # pragma: no cover

    @abstractmethod
    def irecv(self, *, src, tag):
        """Non-blocking one-to-one communication.

        One player (the sender) sends an object to one player (the destination).

        Note
        ----
        Unlike collective operations, this method is only called by the receiver.
        It must be matched by a call to :meth:`send` by the sender.

        See Also
        --------
        recv
            Blocking one-to-one communication.

        Parameters
        ----------
        src: :class:`int`, required
            Rank of the sending player.
        tag: :class:`int` or :class:`~cicada.communicator.interface.Tag`, required
            User- or library-defined tag identifying the message type to match.

        Returns
        -------
        result: :class:`object`
            A special result object that can be used to wait for and access the
            value sent by the sender.  The result object will have a property
            `is_completed` which returns a boolean value indicating whether the
            result has been received; method `wait`, which will block
            indefinitely until the result is received; and property `value`
            which returns the received value or raises an exception if the
            value has not been received yet.
        """
        pass # pragma: no cover

    @abstractmethod
    def isend(self, *, value, dst, tag):
        """Non-blocking one-to-one communication.

        One player (the sender) sends an object to one player (the destination).

        Note
        ----
        Unlike collective operations, this method is only called by the sender.
        It must be matched by a call to :meth:`recv` by the destination.

        See Also
        --------
        send
            Blocking one-to-one communication.

        Parameters
        ----------
        value: Picklable :class:`object`, required
            Value to be sent.
        dst: :class:`int`, required
            Rank of the destination player.
        tag: :class:`int` or :class:`~cicada.communicator.interface.Tag`, required
            User- or library-defined tag identifying the message type.

        Returns
        -------
        result: :class:`object`
            A special result object that can be used to wait until the message
            has been sent.  The result object will have a property
            `is_completed` which returns a boolean value indicating whether the
            result has been sent; and a method `wait` which will block until
            the message is sent.
        """
        pass # pragma: no cover

    @property
    @abstractmethod
    def rank(self):
        """Rank of the local player.

        Returns
        -------
        rank: :class:`int`
            Player rank, in the range :math:`[0, \\text{world_size})`.
        """
        pass # pragma: no cover

    @property
    def ranks(self):
        """List of all ranks managed by this communicator.

        Returns
        -------
        ranks: sequence of :class:`int`
            The set of all ranks managed by this communicator.
        """
        return list(range(self.world_size))


    @abstractmethod
    def recv(self, *, src, tag):
        """Blocking one-to-one communication.

        One player (the sender) sends an object to one player (the destination).

        Note
        ----
        Unlike collective operations, this method is only called by the receiver.
        It must be matched by a call to :meth:`send` by the sender.

        See Also
        --------
        irecv
            Non-blocking one-to-one communication.

        Parameters
        ----------
        src: :class:`int`, required
            Rank of the sending player.
        tag: :class:`int` or :class:`~cicada.communicator.interface.Tag`, required
            User- or library-defined tag identifying the message type to match.

        Returns
        -------
        value: :class:`object`
            The value sent by the sender.
        """
        pass # pragma: no cover

    @abstractmethod
    def scatter(self, *, src, values):
        """One-to-all communication.

        One player (the sender) sends a different object to every player.

        Note
        ----
        This method is a collective operation that *must* be called
        by all players that are members of the communicator.

        Parameters
        ----------
        src: :class:`int`, required
            Rank of the sending player.
        values: sequence of picklable :class:`object`, or `None`, required
            Collection of objects to be sent, one per player, in rank order.

        Returns
        -------
        value: :class:`object`
            The object received by this player.
        """
        pass # pragma: no cover

    @abstractmethod
    def scatterv(self, *, src, values, dst):
        """One-to-many communication.

        One player (the sender) sends a different object to each in a subset of players.

        Note
        ----
        This method is a collective operation that *must* be called
        by all players that are members of the communicator.

        Parameters
        ----------
        src: :class:`int`, required
            Rank of the sending player.
        values: sequence of picklable :class:`object`, or `None`, required
            Collection of objects to be sent, one per recipient.
        dst: sequence of :class:`int`, required
            Rank of each player receiving an object, in the same order as `values`.

        Returns
        -------
        value: :class:`object` or None
            The object received by this player, or `None` if this player wasn't
            in the list of recipients.
        """
        pass # pragma: no cover

    @abstractmethod
    def send(self, value, dst, tag):
        """Blocking one-to-one communication.

        One player (the sender) sends an object to one player (the destination).

        Note
        ----
        Unlike collective operations, this method is only called by the sender.
        It must be matched by a call to :meth:`recv` by the destination.

        See Also
        --------
        isend
            Non-blocking one-to-one communication.

        Parameters
        ----------
        value: Picklable :class:`object`, required
            Value to be sent.
        dst: :class:`int`, required
            Rank of the destination player.
        tag: :class:`int` or :class:`~cicada.communicator.interface.Tag`, required
            User- or library-defined tag identifying the message type.
        """
        pass # pragma: no cover

    @property
    @abstractmethod
    def world_size(self):
        """Number of players sharing this communicator.

        Returns
        -------
        world_size: :class:`int`
            The number of players sharing this communicator.
        """
        pass # pragma: no cover


class Tag(enum.IntEnum):
    """Message tags used internally by the library.

    Callers can use these tags, or any other :class:`int`, when calling
    :meth:`Communicator.send`, :meth:`Communicator.recv`,
    :meth:`Communicator.isend`, and :meth:`Communicator.irecv`.  Note that
    negative integers are reserved for use by the library.
    """

    # Collective operations.
    ALLGATHER = -1
    BARRIER = -2
    BEACON = -3
    BROADCAST = -4
    GATHER = -5
    GATHERV = -6
    REVOKE = -7
    SCATTER = -8
    SCATTERV = -9
    SHRINK = -10

    # Logging operations.
    LOGSYNC = -20

    # Protocol-specific operations.
    PRSZ = -30


def tagname(tag):
    """Return a human-readable name for a tag.

    Parameters
    ----------
    tag: :class:`int` or :class:`~cicada.communicator.interface.Tag`, required
    """
    try:
        tag = Tag(tag)
        return tag.name
    except:
        return str(tag)





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for communicating using the builtin :mod:`socket` module.
"""

import collections
import contextlib
import hashlib
import logging
import multiprocessing
import numbers
import os
import pickle
import queue
import select
import socket
import ssl
import tempfile
import threading
import time
import traceback
import urllib.parse

import numpy
import pynetstring

from ..interface import Communicator, Tag, tagname
from .connect import LoggerAdapter, NetstringSocket, Timeout, Timer, direct, gettls, geturl, listen, message, rendezvous


class BrokenPipe(Exception):
    """Raised trying to send to another player that no longer exists."""
    pass


class Failed(Exception):
    """Used to indicate that a player process raised an exception."""
    def __init__(self, exception, traceback):
        self.exception = exception
        self.traceback = traceback

    def __repr__(self):
        return f"Failed(exception={self.exception!r})" # pragma: no cover


class NotRunning(Exception):
    """Raised calling an operation after the communicator has been freed."""
    pass


class Revoked(Exception):
    """Raised calling an operation after the communicator has been revoked."""
    pass


class Terminated(Exception):
    """Used to indicate that a player process terminated unexpectedly without output."""
    def __init__(self, exitcode):
        self.exitcode = exitcode

    def __repr__(self):
        return f"Terminated(exitcode={self.exitcode!r})" # pragma: no cover


class TryAgain(Exception):
    """Raised when a non-blocking operation would block."""
    pass


class SocketCommunicator(Communicator):
    """Cicada communicator that uses Python's builtin :mod:`socket` module as the transport layer.

    Note
    ----
    Creating a communicator is a collective operation that must be called by
    all players that will be members.

    Parameters
    ----------
    sockets: :class:`dict` of :class:`~cicada.communicator.socket.connect.NetstringSocket`, required
        Dictionary containing sockets that are connected to the other players
        and ready to use.  The dictionary keys must be the ranks of the other
        players, and there must be one socket in the dictionary for every
        player except the caller (since players don't need a socket to
        communicate with themselves).  Note that the communicator world size is
        inferred from the size of the dictionary, and the communicator rank
        from whichever key doesn't appear in the dictionary.
    name: :class:`str`, optional
        Human-readable name for this communicator, used for logging and
        debugging.  Defaults to "world"
    timeout: :class:`numbers.Number`
        Maximum time to wait for communication to complete, in seconds.
    """

    def __init__(self, *, sockets, name="world", timeout=5):
        if not isinstance(sockets, dict):
            raise ValueError("sockets must be a dict, got {sockets} instead.") # pragma: no cover
        for key, socket in sockets.items():
            if not isinstance(key, int):
                raise ValueError("sockets keys must be ints, got {sockets} instead.") # pragma: no cover
            if not isinstance(socket, NetstringSocket):
                raise ValueError("sockets values must be NetstringSocket, got {sockets} instead.") # pragma: no cover

        world_size = len(sockets) + 1
        for index in range(world_size):
            if index not in sockets:
                rank = index

        if not isinstance(name, str):
            raise ValueError("name must be a string, got {name} instead.") # pragma: no cover

        if not isinstance(timeout, numbers.Number):
            raise ValueError(f"timeout must be a number, got {timeout} instead.") # pragma: no cover

        # Setup internal state.
        self._name = name
        self._world_size = world_size
        self._rank = rank
        self._timeout = timeout
        self._revoked = False
        self._log = LoggerAdapter(logging.getLogger(__name__), name, rank)
        self._players = sockets

        self._sent = {}
        self._received = {}

        # Begin normal operation.
        self._running = True

        # Setup queues for incoming messages.
        self._incoming_queue = queue.SimpleQueue()
        self._message_queues = [[] for rank in range(self.world_size)]
        self._message_queue_lock = threading.Lock()

        # Start queueing incoming messages.
        self._queueing_thread = threading.Thread(name=f"Comm {name} player {rank} queueing thread", target=self._queue_messages, daemon=True)
        self._queueing_thread.start()

        # Start receiving incoming messages.
        self._incoming_thread = threading.Thread(name=f"Comm {name} player {rank} incoming thread", target=self._receive_messages, daemon=True)
        self._incoming_thread.start()

        self._log.info(f"communicator ready.")


    def _queue_messages(self):
        # Place incoming messages in the correct queue.
        while self._running:
            # Wait for the next incoming message.
            try:
                raw_message = self._incoming_queue.get(block=True, timeout=0.1)
            except queue.Empty:
                continue

            # Drop messages with missing attributes or unexpected values.
            try:
                src, tag, payload = raw_message
            except Exception as e: # pragma: no cover
                self._log.warning(f"ignoring message: {raw_message} exception: {e}")
                continue

            # Log queued messages.
            if self._log.isEnabledFor(logging.DEBUG):
                self._log.debug(f"<-- player {src} {tagname(tag)}") # pragma: no cover

            try:
                tag = Tag(tag)
            except:
                pass

            if tag not in self._received:
                self._received[tag] = {"messages": 0}
            self._received[tag]["messages"] += 1

            # Revoke messages don't get queued because they receive special handling.
            if tag == Tag.REVOKE:
                if not self._revoked:
                    self._revoked = True
                    self._log.warning(f"revoked by player {src}")
                continue

            # Insert the message into the correct queue.
            with self._message_queue_lock:
                self._message_queues[src].append(raw_message)

        self._log.debug(f"queueing thread closed.")


    def _receive_messages(self):
        # Parse and queue incoming messages as they arrive.
        while self._running:
            # Wait for data to arrive from the other players.
            ready, _, _ = select.select(self._players.values(), [], [], 0.01)
            for src, player in self._players.items():
                if player in ready:
                    try:
                        player.feed()
                    except ConnectionResetError as e: # pragma: no cover
                        # These are pretty common, log them at a lower priority to streamline outputs.
                        self._log.info(f"exception reading from player {src} socket: {e}")
                    except Exception as e: # pragma: no cover
                        self._log.warning(f"exception reading from player {src} socket: {e}")

            # Process any messages that were received. Note that
            # we iterate over every player, not just the ones that
            # were selected above, because there might be a few
            # messages left from the startup process.
            for src, player in self._players.items():
                for raw_message in player.messages():
                    # Ignore unparsable messages.
                    try:
                        tag, payload = pickle.loads(raw_message)
                    except Exception as e: # pragma: no cover
                        self._log.warning(f"ignoring unparsable message: {e}")
                        continue

                    #self._log.debug(f"received {tagname(tag), payload}")

                    # Insert the message into the incoming queue.
                    self._incoming_queue.put((src, tag, payload), block=True, timeout=None)

        # The communicator has been freed, so exit the thread.
        self._log.debug(f"receive thread closed.")


    def _messages(self, *, src=None, tag=None):
        """Return every message that matches the given src and tag.

        Parameters
        ----------
        src: :class:`int` or sequence of :class:`int`, optional
            Return messages from the given player(s).  If :any:`None` (the
            default), return matching messages from every player.
        tag: :class:`int`, optional
            Return messages that match the given tag.  If :any:`None` (the
            default), return messages with any tag.

        Returns
        -------
        messages: :class:`list` of (src, tag, payload) tuples.
        """
        if src is None:
            src = self.ranks
        if not isinstance(src, list):
            src = [src] # pragma: no cover

        matches = []
        with self._message_queue_lock:
            for rank in src:
                misses = []
                for message in self._message_queues[rank]:
                    if tag is not None and tag != message[1]:
                        misses.append(message)
                    else:
                        matches.append(message)
                self._message_queues[rank] = misses
        return matches


    def _next_message(self, *, src, tag):
        """Return the next message (if any) matching the given src and tag.

        Parameters
        ----------
        src: :class:`int`, required
            Return the next matching message from the given player.
        tag: :class:`int`, required
            Return the next matching message with the given tag.
        """
        with self._message_queue_lock:
            for index, message in enumerate(self._message_queues[src]):
                if tag == message[1]: # tag
                    del self._message_queues[src][index]
                    return message
        return None


    def _wait_next_payload(self, *, src, tag):
        """Return the next payload matching the given src and tag, blocking if necessary.

        Parameters
        ----------
        src: :class:`int`, required
            Return the next matching message from the given player.
        tag: :class:`int`, required
            Return the next matching message with the given tag.
        """
        timer = Timer(threshold=self._timeout)
        while not timer.expired:
            message = self._next_message(src=src, tag=tag)
            if message is not None:
                return message[2] # payload
            time.sleep(0.0001)
        raise Timeout(f"Tag {tagname(tag)} from player {src} timed-out after {self._timeout}s")


    def _require_rank(self, rank):
        if not isinstance(rank, numbers.Integral):
            raise ValueError("Rank must be an integer.") # pragma: no cover
        if rank < 0 or rank >= self._world_size:
            raise ValueError(f"Rank must be in the range [0, {self._world_size}).") # pragma: no cover
        return int(rank)


    def _require_rank_list(self, ranks):
        ranks = [self._require_rank(rank) for rank in ranks]
        if len(ranks) != len(set(ranks)):
            raise ValueError("Duplicate ranks are not allowed.") # pragma: no cover
        return ranks


    def _require_running(self):
        if not self._running:
            raise NotRunning(f"Comm {self.name} player {self.rank} is not running.")


    def _require_unrevoked(self):
        if self._revoked:
            raise Revoked(f"Comm {self.name} player {self.rank} has been revoked.")


    def _send(self, *, tag, payload, dst):
        if dst not in self.ranks:
            raise ValueError(f"Unknown destination: {dst}") # pragma: no cover

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug(f"--> player {dst} {tagname(tag)}") # pragma: no cover

        if tag not in self._sent:
            self._sent[tag] = {"messages": 0}
        self._sent[tag]["messages"] += 1

        # As a special-case, route messages sent to ourself directly to the incoming queue.
        if dst == self.rank:
            self._incoming_queue.put((self.rank, tag, payload), block=True, timeout=None)
        # Otherwise, send the message to the appropriate socket.
        else:
            try:
                raw_message = pickle.dumps((int(tag), payload))
                player = self._players[dst]
                player.send(raw_message)
            except BlockingIOError as e: # pragma: no cover
                raise TryAgain(message(self.name, self.rank, f"operation would block sending to player {dst}."))
            except BrokenPipeError as e: # pragma: no cover
                raise BrokenPipe(message(self.name, self.rank, f"broken pipe sending to player {dst}."))


    def allgather(self, value):
        self._log.debug(f"allgather()")

        self._require_unrevoked()
        self._require_running()

        # Send messages.
        for rank in self.ranks:
            self._send(tag=Tag.ALLGATHER, payload=value, dst=rank)

        # Receive messages.
        values = [self._wait_next_payload(src=rank, tag=Tag.ALLGATHER) for rank in self.ranks]
        return values


    def barrier(self):
        """If the implementation returns without raising an exception, then
        every player entered the barrier.  If an exception is raised then there
        are no guarantees about whether every player entered.
        """
        self._log.debug(f"barrier()")

        self._require_unrevoked()
        self._require_running()

        # Notify rank 0 that we've entered the barrier.
        self._send(tag=Tag.BARRIER, payload=None, dst=0)

        if self.rank == 0:
            # Wait until every player enters the barrier.
            for rank in self.ranks:
                self._wait_next_payload(src=rank, tag=Tag.BARRIER)
            # Notify every player that it's time to exit the barrier.
            for rank in self.ranks:
                self._send(tag=Tag.BARRIER, payload=None, dst=rank)

        # Wait until we're told to exit.
        self._wait_next_payload(src=0, tag=Tag.BARRIER)


    def broadcast(self, *, src, value):
        self._log.debug(f"broadcast(src={src})")

        self._require_unrevoked()
        self._require_running()

        src = self._require_rank(src)

        # Broadcast the value to all players.
        if self.rank == src:
            for rank in self.ranks:
                self._send(tag=Tag.BROADCAST, payload=value, dst=rank)

        # Receive the broadcast value.
        return self._wait_next_payload(src=src, tag=Tag.BROADCAST)


    @staticmethod
    def connect(*, world_size=None, rank=None, address=None, root_address=None, identity=None, trusted=None, name="world", timeout=5, startup_timeout=5):
        """High level function to create a SocketCommunicator.

        This is a high level convenience function that can be used to create a
        communicator, given just the calling player's address and the address
        of the root player.  By default, the parameters will be read from
        environment variables that can be set permanently by the user, or
        temporarily using the :ref:`cicada` command.

        Parameters
        ----------
        world_size: :class:`int`, optional
            Number of players.  Defaults to the value of the CICADA_WORLD_SIZE
            environment variable, which is automatically set by the :ref:`cicada` command.
        rank: :class:`int`, optional
            Rank of the caller.  Defaults to the value of the CICADA_RANK
            environment variable, which is automatically set by the :ref:`cicada` command.
        address: :class:`str`, optional
            Listening address of the caller.  This must be a URL of the form
            `"tcp://{host}:{port}"` for TCP sockets, or `"file:///path/to/foo"`
            for Unix domain sockets.  Defaults to the value of the
            CICADA_ADDRESS environment variable, which is automatically set
            by the :ref:`cicada` command.
        root_address: :class:`str`, optional
            Listening address of the root (rank 0) player.  This must be a URL
            of the form `"tcp://{host}:{port}"` for TCP sockets, or
            `"file:///path/to/foo"` for Unix domain sockets.  Defaults to the
            value of the CICADA_ROOT_ADDRESS environment variable, which is
            automatically set by the :ref:`cicada` command.
        identity: :class:`str`, optional
            Path to a private key and certificate in PEM format.  Defaults to
            the value of the CICADA_IDENTITY environment variable, which is
            automatically set by the :ref:`cicada` command.
        trusted: sequence of :class:`str`, optional
            Path to certificates in PEM format.  Defaults to
            the value of the CICADA_TRUSTED environment variable, which is
            automatically set by the :ref:`cicada` command.
        name: :class:`str`, optional
            Human-readable name for the new communicator.  Defaults to "world".
        timeout: :class:`numbers.Number`
            Communication timeout for the new communicator, in seconds.  Defaults to five.
        startup_timeout: :class:`numbers.Number`
            Maximum time to wait for communicator setup, in seconds.  Defaults to five.

        Raises
        ------
        :class:`ValueError`
            If there are problems with input parameters.
        :class:`Timeout`
            If `timeout` seconds elapses before all connections are established.
        :class:`TokenMismatch`
            If every player doesn't provide the same token during startup.

        Returns
        -------
        communicator: :class:`SocketCommunicator`
            A fully-initialized communicator, ready for use.

        """
        if world_size is None:
            world_size = int(os.environ.get("CICADA_WORLD_SIZE"))
        if rank is None:
            rank = int(os.environ.get("CICADA_RANK"))
        if address is None:
            address = os.environ.get("CICADA_ADDRESS")
        if root_address is None:
            root_address = os.environ.get("CICADA_ROOT_ADDRESS")
        if identity is None:
            identity = os.environ.get("CICADA_IDENTITY", "")
        if trusted is None:
            trusted = [trust for trust in os.environ.get("CICADA_TRUSTED", "").split(",") if trust]

        tls = gettls(identity=identity, trusted=trusted)

        timer = Timer(threshold=startup_timeout)
        listen_socket = listen(address=address, rank=rank, name=name, timer=timer)
        sockets = rendezvous(listen_socket=listen_socket, root_address=root_address, world_size=world_size, rank=rank, timer=timer, tls=tls)
        return SocketCommunicator(sockets=sockets, timeout=timeout)


    def free(self):
        # Calling free() multiple times is a no-op.
        if not self._running:
            return

        self._running = False

        # Stop receiving incoming messages.
        self._incoming_thread.join()

        # Stop queueing incoming messages.
        self._queueing_thread.join()

        # Close connections to the other players.
        for player in self._players.values():
            player.close()

        self._log.info(f"communicator freed.")


    def gather(self, *, value, dst):
        self._log.debug(f"gather(dst={dst})")

        self._require_unrevoked()
        self._require_running()

        dst = self._require_rank(dst)

        # Send local data to the destination.
        self._send(tag=Tag.GATHER, payload=value, dst=dst)

        # Receive data from all ranks.
        if self.rank == dst:
            # Messages could arrive out of order.
            values = [self._wait_next_payload(src=rank, tag=Tag.GATHER) for rank in self.ranks]
            return values

        return None


    def gatherv(self, *, src, value, dst):
        self._log.debug(f"gatherv(src={src}, dst={dst})")

        self._require_unrevoked()
        self._require_running()

        src = self._require_rank_list(src)
        dst = self._require_rank(dst)

        # Send local data to the destination.
        if self.rank in src:
            self._send(tag=Tag.GATHERV, payload=value, dst=dst)

        # Receive data from the other players.
        if self.rank == dst:
            # Messages could arrive out of order.
            values = [self._wait_next_payload(src=rank, tag=Tag.GATHERV) for rank in src]
            return values

        return None


    def irecv(self, *, src, tag):
        self._log.debug(f"irecv(src={src}, tag={tagname(tag)})")

        self._require_unrevoked()
        self._require_running()

        src = self._require_rank(src)

        class Result:
            def __init__(self, *, communicator, src, tag):
                self._communicator = communicator
                self._src = src
                self._tag = tag
                self._payload = None

            @property
            def is_completed(self):
                if self._payload is None:
                    message = self._communicator._next_message(src=self._src, tag=self._tag)
                    if message is not None:
                        src, tag, payload = message
                        self._payload = payload
                return self._payload is not None

            @property
            def value(self):
                if self._payload is None:
                    raise RuntimeError("Call not completed.") # pragma: no cover
                return self._payload

            def wait(self):
                if self._payload is None:
                    self._payload = self._communicator._wait_next_payload(src=self._src, tag=self._tag)

        return Result(communicator=self, src=src, tag=tag)


    def isend(self, *, value, dst, tag):
        self._log.debug(f"isend(dst={dst}, tag={tagname(tag)})")

        self._require_unrevoked()
        self._require_running()

        dst = self._require_rank(dst)

        self._send(tag=tag, payload=value, dst=dst)

        # This is safe because we pickle the value before returning; thus,
        # nothing the caller can do to the value will have unexpected
        # side-effects.
        class Result:
            @property
            def is_completed(self):
                return True
            def wait(self):
                pass

        return Result()


    @property
    def name(self):
        """The name of this communicator, which can be used for logging / debugging.

        Returns
        -------
        name: :class:`str`
        """
        return self._name


    @contextlib.contextmanager
    def override(self, *, timeout=None):
        """Temporarily change communicator properties.

        Use :meth:`override` to temporarily modify communicator behavior in a with statement::

            with communicator.override(timeout=10):
                # Do stuff with the new timeout here.
            # The timeout will return to its previous value here.

        Parameters
        ----------
        timeout: :class:`numbers.Number`, optional
            If specified, override the maximum time for communications to complete, in seconds.

        Returns
        -------
        context: :class:`object`
            A context manager object that will restore the communicator state when exited.
        """
        original_context = {
            "timeout": self._timeout,
        }

        try:
            if timeout is not None:
                self._timeout = timeout
            yield original_context
        finally:
            if timeout is not None:
                self._timeout = original_context["timeout"]


    @property
    def rank(self):
        return self._rank


    def recv(self, *, src, tag):
        self._log.debug(f"recv(src={src}, tag={tagname(tag)})")

        self._require_unrevoked()
        self._require_running()

        src = self._require_rank(src)

        return self._wait_next_payload(src=src, tag=tag)


    def revoke(self):
        """Revoke the current communicator.

        Revokes the communicator for this player, and any players able to
        receive messages.  A revoked communicator cannot be used to perform any
        operation other than :meth:`shrink`.  Typically, revoke should be
        called by any player that detects a communication failure, to initiate
        a recovery phase.
        """
        self._log.debug(f"revoke()")

        self._require_running()

        # Notify all players that the communicator is revoked.
        for rank in self.ranks:
            try:
                self._send(tag=Tag.REVOKE, payload=None, dst=rank)
            # Ignore BrokenPipe errors, they're to be expected under the circumstances.
            except BrokenPipe as e: # pragma: no cover
                pass


    @staticmethod
    def run(*, world_size, fn, identities=None, trusted=None, args=(), kwargs={}, family="tcp", name="world", timeout=5, startup_timeout=5):
        """Run a function in parallel using sub-processes on the local host.

        This method returns when the callback functions finish, returning a
        :class:`list` of results from each, in rank order.  Special sentinel
        classes are used to indicate whether a process raised an exception or
        terminated unexpectedly.  This is extremely useful for running examples
        and regression tests on one machine.

        The given function *must* accept a communicator as its first argument.
        Additional caller-provided positional and keyword arguments are passed
        to the function after the communicator.

        To run computation using multiple hosts, you should use
        :meth:`~cicada.communicator.socket.SocketCommunicator.connect` and the
        :ref:`cicada` command line executable instead.

        Parameters
        ----------
        world_size: :class:`int`, required
            The number of players that will run the function.
        fn: :func:`callable`, required
            The function to execute in parallel.
        identities: sequence of :class:`str`, optional
            Path to files in PEM format each containing a private key and a certificate, one per player.
        trusted: sequence of :class:`str`, optional
            Path to files in PEM format containing certificates.
        args: :class:`tuple`, optional
            Positional arguments to pass to `fn` when it is executed.
        kwargs: :class:`dict`, optional
            Keyword arguments to pass to `fn` when it is executed.
        family: :class:`str`, optional
            Address family that matches the scheme used in address URLs
            elsewhere in the API.  Allowed values are "tcp" and "file".
        name: :class:`str`, optional
            Human-readable name for the communicator created by this function.
            Defaults to "world".
        timeout: :class:`numbers.Number`, optional
            Maximum time to wait for normal communication to complete in
            seconds.  Defaults to five seconds.
        startup_timeout: :class:`numbers.Number`, optional
            Maximum time allowed to setup the communicator in seconds.
            Defaults to five seconds.

        Returns
        -------
        results: :class:`list`
            The return value from the function for each player, in rank order.
            Returned only if `return_results` is :any:`True`. If a player
            process terminates unexpectedly, the result will be an instance of
            :class:`Terminated`, which can be used to access the process exit
            code.  If the player process raises a Python exception, the result
            will be an instance of :class:`Failed`, which can be used to access
            the Python exception and a traceback of the failing code.
        """
        def launch(*, parent_queue, child_queue, rank, fn, identity, trusted, args, kwargs, family, name, timeout, startup_timeout):
            # Run the work function.
            try:
                # Create a socket with a randomly-assigned address.
                if family == "file":
                    fd, path = tempfile.mkstemp()
                    os.close(fd)
                    address = f"file://{path}"
                elif family == "tcp":
                    address = "tcp://127.0.0.1"
                timer = Timer(threshold=startup_timeout)
                listen_socket = listen(name=name, rank=rank, address=address, timer=timer)
                address = geturl(listen_socket)

                # Send our address to the parent process.
                parent_queue.put((rank, address))

                # Get all addresses from the parent process.
                addresses = child_queue.get()

                # Optionally setup TLS.
                tls = gettls(identity=identity, trusted=trusted)
                sockets=direct(listen_socket=listen_socket, addresses=addresses, rank=rank, name=name, timer=timer, tls=tls)
                communicator = SocketCommunicator(sockets=sockets, name=name, timeout=timeout)
                result = fn(communicator, *args, **kwargs)
                communicator.free()
            except Exception as e: # pragma: no cover
                result = Failed(e, traceback.format_exc())

            # Return results to the parent process.
            parent_queue.put((rank, result))


        # Setup the multiprocessing context.
        context = multiprocessing.get_context(method="fork") # I don't remember why we prefer fork().

        # Create queues for IPC.
        parent_queue = context.Queue()
        child_queue = context.Queue()

        # Create per-player processes.
        processes = []
        for rank in range(world_size):
            identity = None if identities is None else identities[rank]
            processes.append(context.Process(
                name=f"Player {rank}",
                target=launch,
                kwargs=dict(parent_queue=parent_queue, child_queue=child_queue, rank=rank, fn=fn, identity=identity, trusted=trusted, args=args, family=family, name=name, kwargs=kwargs, timeout=timeout, startup_timeout=startup_timeout),
                ))

        # Start per-player processes.
        for process in processes:
            process.daemon = True
            process.start()

        # Collect addresses from every process.
        addresses = [None] * world_size
        for process in processes:
            rank, address = parent_queue.get(block=True)
            addresses[rank] = address

        # Send addresses to every process.
        for process in processes:
            child_queue.put(addresses)

        # Collect results from processes until every process has completed.
        results = []
        while any([process.is_alive() for process in processes]):
            while True:
                try:
                    rank, result = parent_queue.get(block=False)
                    results.append((rank, result))
                except:
                    break

            time.sleep(0.01)

        # Join all processes, just to be safe.
        for process in processes:
            process.join()

        # Now that every process has exited, collect any remaining results.
        while True:
            try:
                rank, result = parent_queue.get(block=False)
                results.append((rank, result))
            except:
                break

        # Return the output of each rank, in rank order, with a sentinel object for missing outputs.
        output = [Terminated(process.exitcode) for process in processes]
        for rank, result in results:
            output[rank] = result

        # Log the results for each player.
        log = logging.getLogger(__name__)

        for rank, result in enumerate(output):
            if isinstance(result, Failed):
                log.warning(f"Comm {name} player {rank} failed: {result.exception!r}")
            elif isinstance(result, Exception):
                log.warning(f"Comm {name} player {rank} failed: {result!r}")
            else:
                log.info(f"Comm {name} player {rank} result: {result}")

        # Print a traceback for players that failed.
        for rank, result in enumerate(output):
            if isinstance(result, Failed):
                log.error("*" * 80)
                log.error(f"Comm {name} player {rank} traceback:")
                log.error(result.traceback)

        return output


    @staticmethod
    def run_forever(*, world_size, fn, identities=None, trusted=None, args=(), kwargs={}, family="tcp", name="world", timeout=5, startup_timeout=5):
        """Execute a long-running function in parallel using sub-processes on the local host.

        This method returns immediately after networking has been setup and the
        callback function begins executing, returning a :class:`list` of
        network addresses and a :class:`list` of corresponding processes.  This
        is particularly useful for running "MPC-as-a-service" applications on
        the local machine - the caller can use the addresses to communicate with
        the individual processes.

        The given function *must* accept a listening socket and a communicator
        as its first two arguments.  Additional caller-provided positional and
        keyword arguments are passed to the function after the socket and communicator.

        To run a service with multiple hosts, you should use
        :meth:`~cicada.communicator.socket.SocketCommunicator.connect` and the
        :ref:`cicada` command line executable instead.

        Parameters
        ----------
        world_size: :class:`int`, required
            The number of players that will run the function.
        fn: :func:`callable`, required
            The function to execute in parallel.
        identities: sequence of :class:`str`, optional
            Path to files in PEM format each containing a private key and a certificate, one per player.
        trusted: sequence of :class:`str`, optional
            Path to files in PEM format containing certificates.
        args: :class:`tuple`, optional
            Positional arguments to pass to `fn` when it is executed.
        kwargs: :class:`dict`, optional
            Keyword arguments to pass to `fn` when it is executed.
        family: :class:`str`, optional
            Address family that matches the scheme used in address URLs
            elsewhere in the API.  Allowed values are "tcp" and "file".
        name: :class:`str`, optional
            Human-readable name for the communicator created by this function.
            Defaults to "world".
        timeout: :class:`numbers.Number`, optional
            Maximum time to wait for normal communication to complete in
            seconds.  Defaults to five seconds.
        startup_timeout: :class:`numbers.Number`, optional
            Maximum time allowed to setup the communicator in seconds.
            Defaults to five seconds.

        Returns
        -------
        addresses: :class:`list` of :class:`str`
            A listening address for each player, in rank order.  Returned only
            if `return_results` is :any:`False`.
        processes: :class:`list` of :class:`multiprocessing.Process`
            A listening address for each player, in rank order.  Returned only
            if `return_results` is :any:`False`.
        """
        def launch(*, parent_queue, child_queue, rank, fn, identity, trusted, args, kwargs, family, name, timeout, startup_timeout):
            # Run the work function.
            try:
                # Create a socket with a randomly-assigned address.
                if family == "file":
                    fd, path = tempfile.mkstemp()
                    os.close(fd)
                    address = f"file://{path}"
                elif family == "tcp":
                    address = "tcp://127.0.0.1"
                timer = Timer(threshold=startup_timeout)
                listen_socket = listen(name=name, rank=rank, address=address, timer=timer)
                address = geturl(listen_socket)

                # Send our address to the parent process.
                parent_queue.put((rank, address))

                # Get all addresses from the parent process.
                addresses = child_queue.get()

                # Optionally setup TLS.
                tls = gettls(identity=identity, trusted=trusted)
                sockets=direct(listen_socket=listen_socket, addresses=addresses, rank=rank, name=name, timer=timer, tls=tls)
                communicator = SocketCommunicator(sockets=sockets, name=name, timeout=timeout)
                parent_queue.put("ready")
                result = fn(listen_socket, communicator, *args, **kwargs)
                communicator.free()
            except Exception as e: # pragma: no cover
                result = Failed(e, traceback.format_exc())

        # Setup the multiprocessing context.
        context = multiprocessing.get_context(method="fork") # I don't remember why we prefer fork().

        # Create queues for IPC.
        parent_queue = context.Queue()
        child_queue = context.Queue()

        # Create per-player processes.
        processes = []
        for rank in range(world_size):
            identity = None if identities is None else identities[rank]
            processes.append(context.Process(
                name=f"Player {rank}",
                target=launch,
                kwargs=dict(parent_queue=parent_queue, child_queue=child_queue, rank=rank, fn=fn, identity=identity, trusted=trusted, args=args, family=family, name=name, kwargs=kwargs, timeout=timeout, startup_timeout=startup_timeout),
                ))

        # Start per-player processes.
        for process in processes:
            process.daemon = True
            process.start()

        # Collect addresses from every process.
        addresses = [None] * world_size
        for process in processes:
            rank, address = parent_queue.get(block=True)
            addresses[rank] = address

        # Send addresses to every process.
        for process in processes:
            child_queue.put(addresses)

        # Wait until networking has been established.
        for process in processes:
            parent_queue.get(block=True)

        return addresses, processes


    def scatter(self, *, src, values):
        self._log.debug(f"scatter(src={src})")

        self._require_unrevoked()
        self._require_running()

        src = self._require_rank(src)
        if self.rank == src:
            values = [value for value in values]
            if len(values) != self._world_size:
                raise ValueError(f"Expected {self._world_size} values, received {len(values)}.") # pragma: no cover

        # Send data to every player.
        if self.rank == src:
            for value, rank in zip(values, self.ranks):
                self._send(tag=Tag.SCATTER, payload=value, dst=rank)


        # Receive data from the sender.
        return self._wait_next_payload(src=src, tag=Tag.SCATTER)


    def scatterv(self, *, src, values, dst):
        self._log.debug(f"scatterv(src={src}, dst={dst})")

        self._require_unrevoked()
        self._require_running()

        src = self._require_rank(src)
        dst = self._require_rank_list(dst)

        if self.rank == src:
            values = [value for value in values]
            if len(values) != len(dst):
                raise ValueError("values must contain one value instance for every destination player.") # pragma: no cover

        # Send data to every destination.
        if self.rank == src:
            for value, rank in zip(values, dst):
                self._send(tag=Tag.SCATTERV, payload=value, dst=rank)

        # Receive data from the sender.
        if self.rank in dst:
            return self._wait_next_payload(src=src, tag=Tag.SCATTERV)

        return None


    def send(self, *, value, dst, tag):
        self._log.debug(f"send(dst={dst}, tag={tagname(tag)})")

        self._require_unrevoked()
        self._require_running()

        dst = self._require_rank(dst)

        self._send(tag=tag, payload=value, dst=dst)


    def shrink(self, *, name, identity=None, trusted=None, shrink_timeout=5, startup_timeout=5, timeout=5):
        """Create a new communicator containing surviving players.

        This method should be called as part of a failure-recovery phase by as
        many players as possible (ideally, every player still running).  It
        will attempt to rendezvous with the other players and return a new
        communicator, but the process could fail and raise an exception
        instead.  In that case it is up to the application to decide how to
        proceed.

        Parameters
        ----------
        name: :class:`str`, required
            New communicator name.
        identity: :class:`str`, optional
            Path to a private key and certificate in PEM format that will identify the current player.
        trusted: sequence of :class:`str`, optional
            Path to certificates in PEM format that will identify the other players in the new communicator.
        shrink_timeout: :class:`numbers.Number`, optional
            Maximum amount of time to spend identifying remaining members.
        startup_timeout: :class:`numbers.Number`, optional
            Maximum time to wait for communicator_setup, in seconds.
        timeout: :class:`numbers.Number`, optional
            Maximum time to wait for communication, in seconds.

        Returns
        -------
        communicator: :class:`SocketCommunicator`
            New communicator containing the remaining players.

        oldranks: sequence of :class:`int`
            Previous ranks of the remaining players, in rank order.
            Use this if you need to know the original rank of any member of
            `communicator`.
        """
        self._log.debug(f"shrink()")

        self._require_running()

        ##################################################################################
        # Phase 1: Determine which players are still alive.

        # By default, we assume that no-one is alive.
        remaining_ranks = set()

        tls = gettls(identity=identity, trusted=trusted)

        timer = Timer(threshold=shrink_timeout)
        while not timer.expired:
            # Send beacons to the other players (including ourself).
            for rank in self.ranks:
                try:
                    self._send(tag=Tag.BEACON, payload=None, dst=rank)
                # Ignore broken pipe errors, they're to be expected under the circumstances.
                except BrokenPipe: # pragma: no cover
                    pass

            # Get any received beacons (including our own).
            beacons = self._messages(src=None, tag=Tag.BEACON)

            # If we received a beacon, the player is alive.
            for src, tag, payload in beacons:
                remaining_ranks.add(src)

            # If every player is accounted for, we can terminate early.
            if remaining_ranks == set(self.ranks):
                break

            time.sleep(0.5)

        ##################################################################################
        # Phase 2: Use the list of remaining players to generate new network parameters.

        # Sort the remaining ranks; the lowest rank will become rank 0 in the new communicator.
        remaining_ranks = sorted(list(remaining_ranks))

        # Generate a token based on a hash of the remaining ranks that we can
        # use to ensure that every player is in agreement on who's remaining.
        token = hashlib.sha3_256()
        for rank in remaining_ranks:
            token.update(f"rank-{rank}".encode("utf8"))
        token = token.hexdigest()

        # Generate new connection information.
        world_size=len(remaining_ranks)
        rank = remaining_ranks.index(self.rank)

        # Generate a random new listening address.
        for sock in self._players.values():
            address = urllib.parse.urlparse(geturl(sock))
            break
        if address.scheme == "file":
            fd, path = tempfile.mkstemp()
            os.close(fd)
            address = f"file://{path}"
        elif address.scheme == "tcp":
            address = f"tcp://{address.hostname}"
        else:
            raise ValueError(f"Comm {self.name} player {self.rank} cannot parse unknown address scheme: {address.scheme}") # pragma: no cover

        # Create a new listening socket and update the address to match
        timer = Timer(threshold=startup_timeout)
        listen_socket = listen(address=address, rank=rank, name=name, timer=timer)
        address = geturl(listen_socket)

        ##################################################################################
        # Phase 3: Send new network parameters to the remaining players.

        # Send new connection info to the remaining players.
        if self.rank == remaining_ranks[0]:
            for remaining_rank in remaining_ranks:
                self._send(tag=Tag.SHRINK, payload=address, dst=remaining_rank)
        root_address = self._wait_next_payload(src=remaining_ranks[0], tag=Tag.SHRINK)

        # Return a new communicator.
        sockets=rendezvous(listen_socket=listen_socket, root_address=root_address, world_size=world_size, rank=rank, name=name, token=token, timer=timer, tls=tls)
        return SocketCommunicator(sockets=sockets, name=name, timeout=timeout), remaining_ranks


    def split(self, *, name, identity=None, trusted=None, timeout=5, startup_timeout=5):
        """Return a new communicator with the given name.

        If players specify different names - which can be any :class:`str`
        - then a new communicator will be created for each unique name, with
        those players as members.  If a player supplies a name of `None`,
        they will not be a part of any communicator, and this method will
        return `None`.

        .. note::
            This is a collective operation that *must* be called by every member
            of the communicator, even if they aren't going to be a member of any
            of the resulting groups!

        Parameters
        ----------
        name: :class:`str` or :any:`None`, required
            Communicator name, or `None`.
        identity: :class:`str`, optional
            Path to a private key and certificate in PEM format that will identify the current player.
        trusted: sequence of :class:`str`, optional
            Path to certificates in PEM format that will identify the other players in the new communicator.
        timeout: :class:`numbers.Number`, optional
            Maximum time to wait for communication, in seconds.
        startup_timeout: :class:`numbers.Number`, optional
            Maximum time to wait for communicator setup, in seconds.

        Returns
        -------
        communicator: a new :class:`SocketCommunicator` instance, or `None`
        """
        self._log.debug(f"split(name={name})")

        self._require_unrevoked()
        self._require_running()

        if not isinstance(name, (str, type(None))):
            raise ValueError(f"Comm {self.name} player {self.rank} name must be a string or None, got {name} instead.") # pragma: no cover

        tls = gettls(identity=identity, trusted=trusted)

        timer = Timer(threshold=startup_timeout)

        # Generate a random new listening address.
        address = None
        if name is not None:
            for sock in self._players.values():
                address = urllib.parse.urlparse(geturl(sock))
                break
            if address.scheme == "file":
                fd, path = tempfile.mkstemp()
                os.close(fd)
                address = f"file://{path}"
            elif address.scheme == "tcp":
                address = f"tcp://{address.hostname}"
            else:
                raise ValueError(f"Comm {self.name} player {self.rank} cannot split unknown address scheme: {address.scheme}") # pragma: no cover

        # Create a new listening socket and update the address to match.
        if address is not None:
            listen_socket = listen(address=address, rank=self.rank, name=self.name, timer=timer)
            address = geturl(listen_socket)

        # Send names and addresses to rank 0.
        addresses = self.gather(value=(name, address), dst=0)

        # Compute new communicator parameters.
        players = None
        if self.rank == 0:
            groups = collections.defaultdict(list)
            for rank, (name, address) in enumerate(addresses):
                groups[name].append((rank, address))

            players = []
            for rank, (name, address) in enumerate(addresses):
                group = sorted(groups[name])
                ranks, addresses = zip(*group)
                players.append((name, len(group), ranks.index(rank), addresses))

        # Send new connection info to all players.
        group_name, group_world_size, group_rank, group_addresses = self.scatter(src=0, values=players)

        # Return a new communicator.
        if group_name is not None:
            sockets = direct(listen_socket=listen_socket, addresses=group_addresses, rank=group_rank, name=group_name, timer=timer, tls=tls)
            return SocketCommunicator(sockets=sockets, name=group_name, timeout=timeout)

        return None


    @property
    def stats(self):
        """Nested dict containing communication statistics for logging / debugging."""
        results = {
            "player": {},
            "tag": {},
            "total": { "sent": {"bytes": 0, "messages": 0}, "received": {"bytes": 0, "messages": 0}},
        }

        for tag, sent in self._sent.items():
            if tag not in results["tag"]:
                results["tag"][tag] = {}
            results["tag"][tag]["sent"] = sent

        for tag, received in self._received.items():
            if tag not in results["tag"]:
                results["tag"][tag] = {}
            results["tag"][tag]["received"] = received

        for rank, player in self._players.items():
            stats = player.stats
            results["player"][rank] = stats
            results["total"]["sent"]["bytes"] += stats["sent"]["bytes"]
            results["total"]["sent"]["messages"] += stats["sent"]["messages"]
            results["total"]["received"]["bytes"] += stats["received"]["bytes"]
            results["total"]["received"]["messages"] += stats["received"]["messages"]

        return results


    @property
    def timeout(self):
        """Amount of time allowed for communications to complete, in seconds.

        Returns
        -------
        timeout: :class:`numbers.Number`.
            The timeout in seconds.
        """
        return self._timeout


    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout


    @property
    def world_size(self):
        return self._world_size





# Copyright 2021 National Technology & Engineering Solutions
# of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS,
# the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import numpy

from cicada.communicator import SocketCommunicator
from cicada.additive import AdditiveProtcolSuite
from cicada.shamir import ShamirProtocolSuite
from cicada.active import ActiveProtocolSuite

parser = argparse.ArgumentParser(description="Dot product example.")
parser.add_argument("--seed", type=int, default=1234, help="Random seed. Default: %(default)s")
parser.add_argument("--timeout", type=float, default=10, help="Timeout. Default: %(default)ss")
parser.add_argument("--vector-size", "-v", type=int, default=10, help="Vector size. Default: %(default)s")
parser.add_argument("--world-size", "-n", type=int, default=3, help="Number of players. Default: %(default)s")
parser.add_argument("--protocol", "-p", default='additive', help="Protocol to use for SMC. Options are  \"additive\", \"shamir\", or \"active\". Default: %(default)s")
arguments = parser.parse_args()

def main(communicator):
    if arguments.protocol == 'additive':
        protocol = AdditiveProtcolSuite(communicator)
    elif arguments.protocol == 'shamir':
        t = ((arguments.world_size-1)//2)+1
        protocol = ShamirProtocolSuite(communicator, threshold=t)
    elif arguments.protocol == 'active':
        t = ((arguments.world_size-1)//2)+1
        protocol = ActiveProtocolSuite(communicator, threshold=t)
    else:
        raise ValueError('Invalid argument passed for protocol')
    generator = numpy.random.default_rng(seed=arguments.seed + communicator.rank)

    a = generator.uniform(size=arguments.vector_size)
    b = generator.uniform(size=arguments.vector_size)
    if communicator.rank == 0:
        print("  answer:", numpy.dot(a, b))

    ashare = protocol.share(src=0, secret=a, shape=arguments.vector_size)
    bshare = protocol.share(src=0, secret=b, shape=arguments.vector_size)
    cshare = protocol.dot(ashare, bshare)

    print("revealed:", protocol.reveal(cshare))

SocketCommunicator.run(world_size=arguments.world_size, fn=main, timeout=arguments.timeout)
